#include <unistd.h>
#include <time.h>
#include <syslog.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <stddef.h>

#include "inetd.h"

/*
 * Services provided internally by tinetd.
 */

#define	BUFSIZE	8192

/* Echo service -- echo data back */

static void 
echo_vc(int s, struct servtab * sep)
{
	char buffer[BUFSIZE];
	int i;

	s = 0;

	while ((i = t_rcv(s, buffer, sizeof(buffer), NULL)) > 0)
		t_snd(s, buffer, i, 0);
	/* NOTHING */;
	exit(0);
}

/* Echo service -- echo data back */

static void
echo_dg(int s, struct servtab * sep)
{
	char buffer[BUFSIZE];
	struct t_unitdata *ud = t_alloc(s, T_UNITDATA, T_ADDR|T_OPT);
	int i;

	if (!ud)
		return;

	ud->udata.maxlen = sizeof(buffer);
	ud->udata.buf = buffer;
	i = t_rcvudata(s, ud, NULL);
	if (i >= 0)
		i = t_sndudata(s, ud, 0);
	memset(&ud->udata, 0, sizeof(ud->udata));
	t_free(ud, T_UNITDATA);
}

/* Discard service -- ignore data */

static void
discard_vc(int s, struct servtab * sep)
{
	int ret;
	char buffer[BUFSIZE];

	s = 0;     

	while ((i = t_rcv(s, buffer, sizeof(buffer), NULL)) > 0)
		/* NOTHING */;
	exit(0);
}

/* Discard service -- ignore data */

static void
discard_dg(int s, struct servtab * sep)
{
	struct t_unitdata *ud = t_alloc(s, T_UNITDATA, T_UDATA);

	if (!ud)
		return;
	t_rcvudata(s, ud, NULL);
	t_free(ud, T_UNITDATA);
}

#include <ctype.h>

#define LINESIZ 72
static char ring[128];
static char *endring;

static void 
initring(void)
{
	int i;

	endring = ring;

	for (i = 0; i <= 128; ++i)
		if (isprint(i))
			*endring++ = i;
}

/* Character generator */

static void
chargen_vc(int s, struct servtab * sep)
{
	register char *rs;
	int len;
	char text[LINESIZ+2];

	s = 0;

	if (!endring) {
		initring();
		rs = ring;
	}

	text[LINESIZ] = '\r';
	text[LINESIZ + 1] = '\n';
	for (rs = ring;;) {
		if ((len = endring - rs) >= LINESIZ)
			memcpy(text, rs, LINESIZ);
		else {
			memcpy(text, rs, len);
			memcpy(text + len, ring, LINESIZ - len);
		}
		if (++rs == endring)
			rs = ring;
		if (t_snd(s, text, sizeof(text), 0) != sizeof(text))
			break;
	}
	exit(0);
}

/* Character generator */

static void
chargen_dg(int s, struct servtab * sep)
{
	static char *rs;
	int len, size;
	char text[LINESIZ+2];
	struct t_unitdata *ud = t_alloc(s, T_UNITDATA, T_ADDR|T_OPT);
	int i;

	if (!ud)
		return;

	ud->udata.maxlen = sizeof(text);
	ud->udata.buf = text;

	t_rcvudata(s, ud, NULL);

	if (endring == 0) {
		initring();
		rs = ring;
	}

	if (t_rcvudata(s, ud, NULL) < 0) {
		memset(&ud->udata, 0, sizeof(ud->udata));
		t_free(ud, T_UNITDATA);
		return;
	}

	if ((len = endring - rs) >= LINESIZ)
		memcpy(text, rs, LINESIZ);
	else {
		memcpy(text, rs, len);
		memcpy(text + len, ring, LINESIZ - len);
	}
	if (++rs == endring)
		rs = ring;
	text[LINESIZ] = '\r';
	text[LINESIZ + 1] = '\n';
	ud->udata.len = sizeof(text);
	t_sndudata(s, ud, 0);
	memset(&ud->udata, 0, sizeof(ud->udata));
	t_free(ud, T_UNITDATA);
}

/*
 * Return a machine readable date and time, in the form of the
 * number of seconds since midnight, Jan 1, 1900.  Since gettimeofday
 * returns the number of seconds since midnight, Jan 1, 1970,
 * we must add 2208988800 seconds to this figure to make up for
 * some seventy years Bell Labs was asleep.
 */

static long
machtime(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return (htonl((long)tv.tv_sec + 2208988800U));
}

static void
machtime_vc(int s, struct servtab * sep)
{
	long result;
	result = machtime();
	t_snd(s, (char *)&result, sizeof(result);
}

static void
machtime_dg(int s, struct servtab * sep)
{
	long result;
	struct t_unitdata *ud = t_alloc(s, T_UNITDATA, T_ADDR|T_OPT);
	int i;

	if (!ud)
		return;

	ud->udata.maxlen = sizeof(result);
	ud->udata.buf = (char*)result;

	if (t_rcvudata(s, ud, NULL)>=0) {
		result = machtime();
		ud->udata.len = sizeof(result);
		t_sndudata(s, ud, 0);
	}
	memset(&ud->udata, 0, sizeof(ud->udata));
	t_free(ud, T_UNITDATA);
}

/* Return human-readable time of day */

static void
daytime_vc(int s, struct servtab * sep)
{
	char buffer[256];
	time_t time(), clock;

	clock = time((time_t *) 0);
	sprintf(buffer, "%.24s\r\n", ctime(&clock));
	t_snd(s, buffer, strlen(buffer), 0);
}

/* Return human-readable time of day */

static void
daytime_dg(int s, struct servtab * sep)
{
	time_t clock = time(NULL);
	struct t_unitdata *ud = t_alloc(s, T_UNITDATA, T_ADDR|T_OPT);
	char buffer[256];
	int i;

	if (!ud)
		return;

	ud->udata.maxlen = sizeof(buffer);
	ud->udata.buf = buffer;

	if (t_rcvudata(s, ud, NULL)>=0) {
		sprintf(buffer, "%.24s\r\n", ctime(&clock));
		ud->udata.len = strlen(buffer);
		t_sndudata(s, ud, 0);
	}
	memset(&ud->udata, 0, sizeof(ud->udata));
	t_free(ud, T_UNITDATA);
}

struct biltin biltins[] = {
	/* Echo received data */
{	"echo",		"curcuit_v",	1, 0,	echo_vc,	 },
{	"echo",		"datagram_v",	0, 1,	echo_dg,         },

	/* Internet /dev/null */
{	"discard",	"curcuit_v",	1, 0,	discard_vc,      },
{	"discard",	"datagram_v",	0, 1,	discard_dg,      },

	/* Return 32 bit time since 1970 */
{	"time",		"curcuit_v",	0, 0,	machtime_vc,     },
{	"time",		"datagram_v",	0, 1,	machtime_dg,     },

	/* Return human-readable time */
{	"daytime",	"curcuit_v",	0, 0,	daytime_vc,      },
{	"daytime",	"datagram_v",	0, 1,	daytime_dg,      },

	/* Familiar character generator */
{	"chargen",	"curcuit_v",	1, 0,	chargen_vc,      },
{	"chargen",	"datagram_v",	0, 1,	chargen_dg,      },
{	NULL, }
};
