#include <unistd.h>
#include <malloc.h>
#include <limits.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/wait.h>

#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <string.h>

#include "inetd.h"

#define	TOOMANY		40		/* don't start more than TOOMANY */
#define	CNT_INTVL	60		/* servers in CNT_INTVL sec. */
#define	RETRYTIME	(60*5)		/* retry after bind or server fail */
#define RECYCLETIME	(60*15)

#ifdef DEBUG
int	debug = 0;
#endif

static int	nsock, maxsock, max_inetdsock, open_max;
fd_set		*inetd_fdset;
struct pollfd	*inetd_pollset;

static char * config_name = "/etc/tinetd.conf";

static int run_server(struct servtab*, struct port*, struct pollptr*);

static char need_exit = 0;
static char need_terminate = 0;
static char need_resetup = 0;
static char need_reconfig = 0;
static char need_reap_childs = 0;

static void reap_childs(void);
static void closedown(void) __attribute__ ((noreturn));
static void closedown_with_childs(void) __attribute__ ((noreturn));
static void resetup(void);
static void reconfig(void);

#ifdef not_yet
static void dolog(struct port *, int);
#else
#define dolog(a, b) /* */
#endif

static const char INETDPID[] = "/var/run/tinetd.pid";
struct servtab * servtab;

static void
logpid(void)
{
	FILE *fp;

	if ((fp = fopen(INETDPID, "r")) != NULL) {
	        unsigned pid;
		if (fscanf(fp, "%u", &pid) == 1 && kill(0, pid)==0) {
			logerr("tinetd already runs");
			exit(-1);
	        }
		fclose(fp);
	}

	if ((fp = fopen(INETDPID, "w")) != NULL) {
		fprintf(fp, "%u\n", getpid());
		fclose(fp);
	}
}

void
cleanup(int code)
{
	unlink(INETDPID);
	exit(code);
}


/* Signal handlers */

static 
void
sigsegv_handler(int sig)
{
	struct timeval tv;
	int i;
	struct sigset ss;

	tv.tv_sec = 10;
	tv.tv_usec = 0;
	logerr("signal %d got", sig);
	for (i=0; i < open_max; i++)
		close(i);
	select(0, NULL, NULL, NULL, &tv);
	sigemptyset(&ss);
	alarm(0);
	sigprocmask(SIG_SETMASK, &ss, NULL);
	execl("/sbin/inetd", "inetd", "-d", "2", NULL);
	cleanup(-1);
}

static void
sigdebug_handler(int sig)
{
	debug++;
	if (sig == SIGUSR2)
		debug = 0;
	syslog(LOG_DEBUG, "debuggind level set to %d", debug);
}

static void
sigint_handler(int sig)
{
	need_exit = 1;
}

static void
sigterm_handler(int sig)
{
	need_terminate = 1;
}


static void
sigalrm_handler(int sig)
{
	need_resetup = 1;
}

static void
sighup_handler(int sig)
{
	need_reconfig = 1;
}

static void
sigchld_handler(int sig)
{
	need_reap_childs = 1;
}

static char timingout=0;

static void
set_timeout(void)
{
	if (!timingout) {
		timingout = 1;
		alarm(RETRYTIME);
	}
}

__inline__
static void
reset_timeout(void)
{
	timingout = 0;
	alarm(RECYCLETIME);
}

void
bad_descriptor(int n)
{
	int i;
	struct servtab * sep;
	struct port * port;
	static int errcnt = 0;

	logerr("bad descriptor %d\n", n);

	if (++errcnt > 16) {
		logerr("panic! Too many bad descriptors.\n");
		for (sep = servtab; sep; sep = sep->next) {
			if (sep->wait && sep->pid)
				kill(SIGKILL, sep->pid); 
		}
		sigsegv_handler(0);
		/* DOES NOT RETURN */
		exit(1);
	}
		
	for (sep = servtab; sep; sep = sep->next) {
		for (port = sep->port; port; port = port->next) {
			if (port->fd != n)
				continue;
			logerr("svc %s\n", sep->server);
			if (sep->wait && sep->pid)
				kill(SIGKILL, sep->pid);
			port->fd = -1;
			sep->pid = 0;
			sep->obsolete = 1;
			debug = 1;
			need_resetup = 1;
		}
	}
}

void
main (int argc, char * argv[])
{
#ifdef DEBUG
	_BLOCK_BEGIN
	int ch;

	while ((ch = getopt(argc, argv, "d:")) != EOF) {
		switch(ch) {
		case 'd':
			debug = atoi(optarg);
			break;
		case '?':
		default:
			fprintf(stderr, "usage: inetd [-d level]\n");
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;
	_BLOCK_END
#endif

	if (argc > 0) {
		config_name = argv[0];
		if (debug)
			fprintf(stderr, "config is \"%s\"\n", config_name);
	}

	init_nconf();

	open_max = __rpc_dtbsize();
	inetd_fdset = (fd_set*)malloc((open_max+NFDBITS-1)/8);
	if (inetd_fdset == NULL)
		exit(2);
	memset(inetd_fdset, 0, (open_max+NFDBITS-1)/8);
	inetd_pollset = (struct pollfd*)malloc(open_max*sizeof(struct pollfd));
	if (inetd_pollset == NULL)
		exit(3);
	memset(inetd_pollset, 0, open_max*sizeof(struct pollfd));

	_BLOCK_BEGIN
	extern void daemon(int, int);
	daemon(0, 0);

	_BLOCK_END
	openlog("inetd", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
	logpid();
   
	_BLOCK_BEGIN
	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = sigalrm_handler;
	sigaction(SIGALRM, &sa, NULL);
	sa.sa_handler = sighup_handler;
	sigaction(SIGHUP, &sa, NULL);
	sa.sa_handler = sigchld_handler;
	sigaction(SIGCHLD, &sa, NULL);
	sa.sa_handler = sigterm_handler;
	sigaction(SIGTERM, &sa, NULL);
	sa.sa_handler = sigint_handler;
	sigaction(SIGINT, &sa, NULL);
	sa.sa_handler = sigdebug_handler;
	sigaction(SIGUSR1, &sa, NULL);
	sigaction(SIGUSR2, &sa, NULL);

	sa.sa_handler = sigsegv_handler;
	sigfillset(&sa.sa_mask);
	sigaction(SIGSEGV, &sa, NULL);
	need_reconfig = 1;
	_BLOCK_END

	for (;;) {
		if (need_exit)
			closedown();
		if (need_terminate)
			closedown_with_childs();
		if (need_reap_childs) {
			reap_childs();
			continue;
		}
		if (need_reconfig) {
			reconfig();
			continue;
		}
		if (need_resetup) {
			resetup();
			continue;
		}
		if (nsock == 0) {
			pause();
			continue;
		}

		_BLOCK_BEGIN
		int n;
		extern fd_set *svc_fdset_p;

		n = __rpc_select_to_poll(open_max, svc_fdset_p, inetd_pollset);
		n += __rpc_select_to_poll(max_inetdsock, inetd_fdset, inetd_pollset+n);
		maxsock = n;

		if ((n = poll(n, inetd_pollset, INFTIM)) <= 0) {
			if (n < 0 && errno != EINTR) {
				logerr("poll: %m\n");
				sleep(1);
			}
			continue;
		}

		_BLOCK_BEGIN
		struct servtab *sep;
		struct port *port;
		int i;

		for (i=0; i<n; i++) {
			if (inetd_pollset[i].fd < 0 || inetd_pollset[i].revents==0)
				continue;
			if (FD_ISSET(i, svc_fdset_p)) {
				svc_getreq_poll(inetd_pollset+i, 1);
				continue;
			}
			if (inetd_pollset[i].revents & POLLNVAL) {
				bad_descriptor(i);
				continue;
			}
			for (sep = servtab; sep; sep = sep->next) {
				for (port = sep->port; port; port = port->next) {
					if (port->fd == inetd_pollset[i].fd) {
						dprintf(1, "someone wants [%s/%s]", port->service, port->nconf->nc_netid);
						run_server(sep, port, inetd_pollset+i);
					}
				}
			}
		}
		_BLOCK_END
		_BLOCK_END
	}
}



/* Closedown procedure */

static void
closedown(void)
{
	loginfo("inetd terminated");
	cleanup(0);
}

static void
closedown_with_childs(void)
{
	struct servtab * sep;
	for (sep = servtab; sep; sep = sep->next) {
		if (sep->wait && sep->pid)
			kill(SIGTERM, sep->pid); 
	}
	loginfo("inetd terminated with childs");
	cleanup(0);
}

static struct port * leaved_ports = NULL;

void
activate_fd(int fd)
{
	int  onoff = 0;
	if (fd < 0 || FD_ISSET(fd, inetd_fdset))
		return;
	FD_SET(fd, inetd_fdset);
	nsock++;
	if (fd > max_inetdsock)
		max_inetdsock = fd;
	ioctl(fd, FIONBIO, (char *) &onoff);
}

static void
deactivate_fd(int fd)
{
	if (fd < 0 || !FD_ISSET(fd, inetd_fdset))
		return;
	FD_CLR(fd, inetd_fdset);
	if (fd == max_inetdsock)
		max_inetdsock--;
	nsock--;
}


/* Port list tools */

static void 
remove_from_port_list(struct port ** pport, struct port * port)
{
#ifdef PARANOID
	if (!pport) {
		logerr("par1");
		cleanup(1);
	}
	if (!*pport) {
		logerr("par13");
		cleanup(1);
	}
	if ((*pport)->prev) {
		logerr("par21");
		cleanup(1);
	}
#endif
	if (port->next)
		port->next->prev = port->prev;
	if (port->prev)
		port->prev->next = port->next;
	if (*pport == port)
		*pport = port->next;
}

void 
add_to_port_list(struct port ** pport, struct port * port)
{
#ifdef PARANOID
	if (!pport) {
		logerr("par2");
		cleanup(1);
	}
	if (*pport && (*pport)->prev) {
		logerr("par3");
		cleanup(1);
	}
#endif
	port->next = *pport;
	port->prev = NULL;
	if (*pport)
		(*pport)->prev = port;
	*pport = port;
}

static void 
free_port(struct port * port)
{
#ifdef not_yet
	struct pmaplist * pml=port->pml;
	while (pml) {
		struct pmaplist * nextpml = pml->pml_next;
		free(pml);
		pml = nextpml;
	}
#endif
	free(port);
}

static void 
discard_port_list(struct port ** pport)
{
	struct port * port = *pport;
	struct port * next;
#ifdef PARANOID
	if (!pport) {
		logerr("par4");
		cleanup(1);
	}
#endif
	while (port) {
		next = port->next;
		remove_from_port_list(pport, port);
#ifdef not_yet
		if (port->fd < 0 || (port->pml && port->pml->pml_map.pm_prog == PMAPPROG))
#else
		if (port->fd < 0)
#endif
			free_port(port);
		else
			add_to_port_list(&leaved_ports, port);
		port = next;
	}
}

void 
free_port_list(struct port ** pport)
{
	struct port * port = *pport;
	struct port * next;
#ifdef PARANOID
	if (!pport) {
		logerr("par5");
		cleanup(1);
	}
#endif
	while (port) {
		next = port->next;
		remove_from_port_list(pport, port);
		if (port->fd >= 0) {
			deactivate_fd(port->fd);
			close(port->fd);
		}
		free_port(port);
		port = next;
	}
}

static struct servtab * 
lookup_port(struct port * port)
{
	struct servtab  * sep;
	struct port  * p;

	for (sep = servtab; sep; sep = sep->next) {
		for (p = sep->port; p; p = p->next) {
			if (port->nconf == p->nconf &&
			    port->bnd.addr.len == p->bnd.addr.len &&
			    memcmp(port->bnd.addr.buf, p->bnd.addr.buf, p->bnd.addr.len) == 0)
				return sep;
		}
	}
	return NULL;
}

static struct port *
find_port_in_list(struct port * port, struct port * list)
{
	register struct port  * p;
	for (p = list; p; p = p->next) {
		if (ports_match(port,p))
			return p;
	}
	return NULL;
}


/* servtab list tools */

static void 
free_servtab(struct servtab *cp)
{
	int i;

	free_port_list(&cp->port);
	if (cp->user)
		free(cp->user);
	if (cp->server)
		free(cp->server);
	for (i = 0; i < MAXARGV; i++)
		if (cp->argv[i])
			free(cp->argv[i]);
}

static void 
remove_from_servtab_list(struct servtab ** sepp, struct servtab * sep)
{
#ifdef PARANOID
	if (!sepp) {
		logerr("par6");
		cleanup(1);
	}
	if (!*sepp) {
		logerr("par12");
		cleanup(1);
	}
	if ((*sepp)->prev) {
		logerr("par20");
		cleanup(1);
	}
#endif
	if (sep->next) 
		sep->next->prev = sep->prev;
	if (sep->prev)
		sep->prev->next = sep->next;
	if (*sepp == sep)
		*sepp = sep->next;
}


static void 
add_to_servtab_list(struct servtab ** sepp, struct servtab * sep)
{
#ifdef PARANOID
	if (!sepp) {
		logerr("par7");
		cleanup(1);
	}
	if (*sepp && (*sepp)->prev) {
		logerr("par8");
		cleanup(1);
	}
	if (!sep) {
		logerr("par9");
		cleanup(1);
	}
#endif
	sep->next = *sepp;
	sep->prev = NULL;
	if (*sepp)
		(*sepp)->prev = sep;
	*sepp = sep;
}


void 
discard_servtab(struct servtab * sep)
{
	discard_port_list(&sep->port);
	free_servtab(sep);
}

static void 
discard_servtab_list(struct servtab ** sepp)
{
	struct servtab *sep, *next;

#ifdef PARANOID
	if (!sepp) {
		logerr("par10");
		cleanup(1);
	}
#endif
	sep = *sepp;
	while (sep) {
		next = sep->next;
		if (!sep->wait || sep->pid == 0) {
			remove_from_servtab_list(sepp, sep);
			discard_servtab(sep);
		}
		sep = next;
	}
}


static void 
reconfig(void)
{
	register struct servtab *sep;
	struct port *port;
	time_t curtime = time(NULL);

	dprintf(1,"reconfig");

	if (!setconfig(config_name)) {
		logerr( "cannot open %s: %m", config_name);
		if (!servtab)
			cleanup(1);
		logerr( "reconfig command ignored");
		need_reconfig = 0;
		return;
	}

#ifdef not_yet
	setrpcent(1);
#endif

	discard_servtab_list(&servtab);

	for (sep = servtab; sep; sep = sep->next) {
		sep->obsolete = 1; 
		dprintf(1, "%s still running", sep->server);
	}

#ifdef not_yet
	shutdown_portmap();
#endif

more:

	while ((sep = getconfigent()) != NULL) {
		for (port = sep->port; port; port=port->next) {
			struct servtab * s;

			port->last_use = curtime;
			if (!port->bnd.addr.len && !port->rpcbl) {
				logerr( "a port for %s is not resolved, ignored.", sep->server);
				discard_servtab(sep);
				goto more;
			}
			if (port->bnd.addr.len) {
				if ((s = lookup_port(port)) != NULL) {
					if (!s->obsolete) {
						logerr("port [%s/%s] listened by >1 services", port->service, port->nconf->nc_netid);
						discard_servtab(sep);
						goto more;
					}
					port->fd = -2;
				} else if (find_port_in_list(port,leaved_ports))
					port->fd = -2;
			} else {
#ifdef not_yet
				struct port     * oldport = NULL;
				struct rpcbinding * rpcbl;

				for (rpcbl = port->rpcbl; rpcbl; rpcbl = rpcbl->rpcbl_next) {
					struct port     * tport;
					if (find_rpc_service(pml->pml_map.pm_prog,
							     pml->pml_map.pm_vers,
							     pml->pml_map.pm_prot, &s, &tport)) {
						if (!s->obsolete) {
							logerr( "rpc%d listened by >1 services", pml->pml_map.pm_prog);
							discard_servtab(sep);
							goto more;
						}
						if (IN_PORT(tport))
							oldport = tport;
					} else
						dprintf(1, "rpc%d/%d/%d not found\n", pml->pml_map.pm_prog,
							pml->pml_map.pm_vers,
							pml->pml_map.pm_prot);
				}
				if (oldport) {
					port->fd = -2;
					IN_PORT(port) = IN_PORT(oldport);
					for (pml = port->pml; pml; pml = pml->pml_next) {
						pml->pml_map.pm_port = ntohs(IN_PORT(oldport));
						dprintf(1, "rpc%d/%d/%d waiting for port %d\n",
							pml->pml_map.pm_prog,
							pml->pml_map.pm_vers,
							pml->pml_map.pm_prot,
							pml->pml_map.pm_port);
					}
				}
#endif
			}
		}
		add_to_servtab_list(&servtab, sep);
	}
#ifdef not_yet
	endrpcent();
#endif
	endconfig();
	need_reconfig = 0;
	resetup();
}


void
set_port_options(struct servtab * sep, struct port * port)
{
	int off_on[2] = {0, 1};

#ifdef DEBUG
	if (setsockopt(port->fd, SOL_SOCKET, SO_DEBUG, (char *)&off_on[debug > 1 || port->debug], 
#else
	if (setsockopt(port->fd, SOL_SOCKET, SO_DEBUG, (char *)&off_on[port->debug], 
#endif
		       sizeof (int))<0) {
		logerr( "setsockopt (SO_DEBUG): %m");
	}

	if (setsockopt(port->fd, SOL_SOCKET, SO_REUSEADDR, 
		       (char *)&off_on[port->reuse], sizeof (int))<0) {
		logerr( "setsockopt (SO_REUSEADDR) = %i: %m", port->reuse);
	}
	if (setsockopt(port->fd, SOL_SOCKET, SO_KEEPALIVE, 
		       (char *)&off_on[port->keepalive], sizeof (int))<0) {
		logerr( "setsockopt (SO_KEEPALIVE) = %i: %m", port->keepalive);
        }
}


static void
flush_port(struct port * port, struct pollfd *pollptr, int nfds)
{
	int i;

	if (port->fd < 0)
		return;

	for (i=0; i<nfds; i++) {
		if (pollptr[i].fd != port->fd)
			continue;
		if (!pollptr[i].revents)
			break;
		pollptr[i].revents = 0;
		if (port->nconf->nc_semantics == NC_TPI_CLTS) {
			char buf[4];
			struct t_udata ud;
			memset(&ud, 0, sizeof(ud));
			ud.udata.buf = buf;
			ud.udata.maxlen = sizeof(buf);
			t_rcvudata(port->fd, ud, NULL);
		}
	}
}

static void
flush_dgram_ports(struct pollfd *pollset, int nfds)
{
	int n;
	for (n = 0; n < nfds; n++) {
		if (pollptr[n].fd >=0 && pollptr[n].revents) {
			char buf[4];
			struct t_udata ud;
			pollptr[n].revents = 0;
			memset(&ud, 0, sizeof(ud));
			ud.udata.buf = buf;
			ud.udata.maxlen = sizeof(buf);
			t_rcvudata(port->fd, ud, NULL);
		}
	}
}

static void
deactivate_service(struct servtab * sep)
{
	struct port * port;
	for (port = sep->port; port; port = port->next) 
		deactivate_fd(port->fd);
}

static int
activate_service(struct servtab * sep)
{
	struct port * port;
	time_t curtime = time(NULL);

	for (port = sep->port; port; port = port->next) {
		port->last_use =  curtime;
		if (port->fd < 0) {
			deactivate_service(sep);
			return -1;
		}
	}
#ifdef not_yet
	if (sep->port->pml && sep->port->pml->pml_map.pm_prog == PMAPPROG)
		return resetup_portmap(sep);
#endif
	for (port = sep->port; port; port = port->next) 
		activate_fd(port->fd);
	return 0;
}


static int
open_port(struct servtab * sep, struct port * port)
{
	int fd;
	if ((fd = t_open(port->nconf->nc_device, O_RDWR, NULL)) < 0) {
		logerr( "t_open: [%s/%s]: %m", port->service, post->nconf->nc_netid);
		return -1;
	}
	if (!sep->wait) {
		int on = 1;
		if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, 
			       (char *)&on, sizeof (on))<0)
			logerr( "open_port: setsockopt (SO_REUSEADDR) = 1: %m");
	}
	if (t_bind(fd, &port->bnd, NULL) < 0) {
		logerr( "t_bind: [%s/%s]: %m", port->service, post->nconf->nc_netid);
		t_close(fd);
		set_timeout(); 
		return -1;
	}
	port->fd = fd;
	set_port_options(sep, port);
	activate_fd(fd);
	return fd;
}


#ifdef not_yet
int
bind_rpc_service(struct servtab * sep)
{
  struct port * port;
  struct pmaplist * pml;
  time_t curtime;
  time(&curtime);

  for (port = sep->port; port; port = port->next) {
    int fd;
    u_long prog, prot;
    port->last_use = curtime;

    if (IN_PORT(port))
      continue;
#ifdef PARANOID
    if (!port->pml) {
      logerr("par100");
      cleanup(1);
    }
#endif
    prog = port->pml->pml_map.pm_prog; 
    prot = port->pml->pml_map.pm_prot; 
    if ((fd = socket(FAMILY(port), port->socktype, 0)) < 0) {
      logerr( "socket: rpc%d/%d: %m", prog, prot);
      return 0;
    }
    if (bindresvport(fd, (struct sockaddr_in*)&port->addr) < 0) {
      logerr( "bindresvport: rpc%d/%d: %m", prog, prot);
      close(fd);
      return 0;
    }
    port->fd = fd;
    set_port_options(sep, port);
    for (pml = port->pml; pml; pml=pml->pml_next) {
      dprintf(2, "rpc%d bound to %d at %d", pml->pml_map.pm_prog,
	     ntohs(IN_PORT(port)), fd);
      pml->pml_map.pm_port = ntohs(IN_PORT(port));
    }
    if (port->socktype == SOCK_STREAM)
      listen(fd, 5);
    activate_fd(fd);
  }
  return 1;
}
#endif


static void
resetup()
{
	struct servtab * sep;
	struct port * port;
	time_t curtime = time(NULL);
  
	dprintf(1,"resetup");

	reset_timeout();

#ifdef not_yet
	for (sep = servtab; sep; sep = sep->next) {
		if (sep->obsolete || (sep->wait && sep->pid))
			continue;
		for (port = sep->port; port; port = port->next) {
			if (port->autorpc && IN_PORT(port) && 
			    curtime >= port->last_use + RECYCLETIME*2) {
				struct pmaplist * pml;
#ifdef PARANOID
				if (!port->pml) {
					logerr("par300");
					cleanup(1);
				}
#endif
				dprintf(1, "port%i/%i is recycled\n", ntohs(IN_PORT(port)), port->proto);
				for (pml = port->pml; pml; pml = pml->pml_next)
					pml->pml_map.pm_port = 0;
				if (port->fd >= 0) {
					deactivate_fd(port->fd);
					close(port->fd);
					port->fd = -1;
				}
				IN_PORT(port) = 0;
			}
		}
	}
#endif

	for (sep = servtab; sep; sep = sep->next) {
		int something_done = 0;
		if (sep->obsolete || (sep->wait && sep->pid))
			continue;
		for (port = sep->port; port; port = port->next) {
			switch (-port->fd) {
			case 1:
				if (port->bnd.addr.len) {
					if (open_port(sep, port) >= 0)
						something_done = 1;
					dprintf(1, "opening [%s/%s] -> %i", port->service, port->nconf->nc_netid, port->fd);
				}
				continue;
			case 2:
			{ 
				struct port * p;
				if ((p = find_port_in_list(port, leaved_ports))  !=  NULL) {
					dprintf(1, "restoring [%s/%s] -> %i", port->service, port->nconf->nc_netid, p->fd);
					remove_from_port_list(&leaved_ports, p);
					port->fd = p->fd;
					p->fd = -1;
					something_done = 1;
					free_port(p);
					set_port_options(sep, port);
				} else
					dprintf(1, "restore [%s/%s] failed", port->service, port->nconf->nc_netid);
				continue;
			}
			default:
				continue;
			}
		}
		if (something_done)
			activate_service(sep);
	}
#ifdef DEBUG
	for (port = leaved_ports; port; port = port->next)
		dprintf(1, "port [%s/%s] is cancelled", port->service, port->nconf->nc_netid);
#endif
	free_port_list(&leaved_ports);
	need_resetup = 0;
}


static pid_t 
fork_child(struct servtab *sep, struct port * port, fd_set * readable)
{
	if (sep->count++ == 0)
		gettimeofday(&sep->start_time, NULL);
	else if (sep->count >= TOOMANY) {
		struct timeval now;
		gettimeofday(&now, NULL);
		if (now.tv_sec - sep->start_time.tv_sec > CNT_INTVL) {
			sep->start_time = now;
			sep->count = 1;
		} else {
			logerr( "%s server, service terminated", sep->server);
			sep->count = 0;

			set_timeout();
			errno = 0;
			return -1;
		}
	}
	return fork();
}

/* These two procedures started only after forking,
 * they rearrange file descriptors to pass them to a service.  
 */

static void
fork_accepted_port(struct servtab * sep, int ctrl)
{
	int i;
	dprintf(1, "forked one port");
#ifdef PARANOID
	if (ctrl <= 2) {
		logerr("fork_accepted_port : ctrl<=2");
		return; 
	}  
#endif
	for (i = open_max-1; i > 2; i--)
		if (i != ctrl)
			close(i);
	dup2(ctrl, 0);
	close(ctrl);
	dup2(0, 1);
	dup2(0, 2);
	if (sep->port->downgrade && sep->port->ops &&
	    sep->port->ops->downgrade)
		sep->port->ops->downgrade(sep, port, 0);
}

static void
fork_multi_ports(struct servtab * sep, struct pollfd *pollset, int *nfds)
{
	int n,m;
	int dg_fds = 0;
	dprintf(1, "forked multi port");

	FD_ZERO(readable);
	for (n=0, m=0; n <= max_inetdsock; n++) {
		if (FD_ISSET(n, inetd_fdset)) {
			struct port * p;
			for (p = sep->port; p; p = p->next) {
				if (p->fd == n)
					break;
			}
			if (!p)
				continue;
			dprintf(1, "[%s/%s] socket(%i) dupped to %i", p->service, p->nconf->nc_netid, n, m);
			if (n != m)
				dup2(n, m);
			if (p->nconf->nc_semantics == NC_TPI_CLTS) {
				pollset[dg_fds].fd = m;
				pollset[dg_fds].revents = -1;
				dg_fds++;
			}
			m++;
		}
	}

	if (m < 3) {
		int tfd = open("/dev/null",O_RDWR);
		for (; m < 3 ; m++) {
			if (m != tfd)
				dup2(tfd, m);
		} 
		close(tfd);
	}
	for (; m < open_max; m++)
		close(m);
	*nfds = dg_fds;
}


static int
run_server(struct servtab *sep, struct port *port, struct pollfd *pollptr, int nfds)
{
	struct port * p;
	int ctrl = -1;

	if (sep->wait) { 
		for (p = sep->port; p; p = p->next) {
			if (p->fd < 0) {
				logerr("cannot start %s, port is busy", sep->server);
				deactivate_service(sep);
				for (p = sep->port; p; p = p->next) 
					flush_port(p, pollptr, nfds);
				return -1;
			}
		}
	}

	if (!sep->wait) {
		int onoff = 1;
		switch (port->nconf->nc_semantics) {
		case NC_TPI_COTS:
		case NC_TPI_COTS_ORD:
			ioctl(port->fd, FIONBIO, (char *) &onoff);
			ctrl = accept(port->fd, NULL, NULL);
			onoff = 0;
			ioctl(port->fd, FIONBIO, (char *) &onoff);
			break;
		default:
			ctrl = dup(port->fd);
		}
		if (ctrl < 0) {
			if (errno !=  EINTR && errno != EWOULDBLOCK && errno != EAGAIN)
				logerr( "accept (for %i): %m", ntohs(IN_PORT(port)));
			return 1;
		}
		onoff = 1;
		if (setsockopt(ctrl, SOL_SOCKET, SO_REUSEADDR, 
			       (char *)&onoff, sizeof (onoff))<0) {
			logerr("setsockopt (SO_REUSEADDR) of accepted port: %m");
		}
		dprintf(1, "accepted [%s/%s]", port->service, port->nconf->nc_netid);
	}

	/* No fork needed */

	if (sep->bi && !sep->bi->forking) {
		(*sep->bi->bi_fn)( (ctrl >=0 ? ctrl : port->fd ), sep);
		if (ctrl >= 0)
			close(ctrl);
		return 1;
	};


	_BLOCK_BEGIN
	int pid;
	pid = fork_child(sep, port, pollptr); 

	if (pid < 0) {                                /* Fork failure */
		if (errno) {
			logerr( "fork: %m");
			sleep(1);
		} 

		if (sep->wait) {
			struct port * p;
			deactivate_service(sep);
			for (p = sep->port; p; p = p->next) {
				flush_port(p, pollptr, nfds);
				close(p->fd);
				p->fd = -1;
			}
			return 0;
		} else {
			deactivate_fd(port->fd);
			flush_port(port, pollptr, nfds);
			close(port->fd);
			port->fd = -1;
			if (ctrl >= 0)
				close(ctrl);
			return 1;
		}
	} else if (pid) {                        /* Fork done; parent side */
		if ( sep->wait) {
			struct port * p;
			sep->pid = pid;
			deactivate_service(sep);
			for (p = sep->port; p; p=p->next) {
				if (p->fd >= 0) {
					FD_CLR(p->fd, readable);
				}
			}
			return 0;
		} else {
			if (ctrl >= 0)
				close(ctrl);
			return 1;
		}
	} else {                                      /* Fork done; child side */
		openlog("inetd_child", LOG_PID, LOG_DAEMON);

		if (port->log)
			dolog(port, ctrl >=0 ? ctrl : port->fd);

		if (ctrl >= 0)
			fork_accepted_port(sep, ctrl);
		else
			fork_multi_ports(sep, inetd_pollset, &nfds);

		if (sep->bi) {
			(*sep->bi->bi_fn)(0, sep);
			_exit(0);
		} else {
			register struct passwd *pwd;
			dprintf(1, "%d execl %s\n", getpid(), sep->server);
			if ((pwd = getpwnam(sep->user)) == NULL) {
				logerr( "getpwnam: %s: No such user", sep->user);
				flush_dgram_ports(inetd_pollset, nfds);
				_exit(-1);
			}
			if (pwd->pw_uid) {
				setgid((gid_t)pwd->pw_gid);
				initgroups(pwd->pw_name, pwd->pw_gid);
				setuid((uid_t)pwd->pw_uid);
			}
			execv(sep->server, sep->argv);
			logerr( "execv %s: %m", sep->server);
			flush_dgram_ports(inetd_pollset, nfds);
			_exit(1);
		}
		/*NOT REACHED*/  
	}
	_BLOCK_END
	/*NOT REACHED*/  
}


/* Procedure handling reaped childs */

static void
reap_childs(void)
{
	char * name = "rpc_callit";

	for (;;) {
		int status;
		int pid;
		struct servtab * sep;
		pid = wait3(&status, WNOHANG, NULL);
		if (pid <= 0) {
			need_reap_childs = 0;
			break;
		}
		for (sep = servtab; sep; sep = sep->next) {
			if (sep->wait && sep->pid == pid) {
				name = sep->server;
				break;
			}
		}
#ifdef not_yet
		if (!sep && !reap_rpc_callit(pid))
			continue;
#else
		if (!sep)
			continue;
#endif
		dprintf(1, "%s[%i] reaped", name, pid);
		if (WIFEXITED(status)) {
			if (WEXITSTATUS(status)) {
				logerr("%s: exited with status %i",
				       name, WEXITSTATUS(status));
			}
		} else if (WIFSIGNALED(status)) {
			logerr("%s: got signal \"%s\"",
			       name, sys_siglist[WTERMSIG(status)]);
		} else if (WIFSTOPPED(status)) {
			logerr("%s: stopped by signal \"%s\"",
			       name, sys_siglist[WSTOPSIG(status)]);
			continue;
		} else {
			logerr( "%s: exited with invalid status 0x%08x",
			       name, status);
		}
 		
		if (!sep)
			continue;

		if (sep->obsolete) {
			discard_servtab(sep);
			need_resetup = 1;
		} else {
			sep->pid = 0;
			activate_service(sep);
		}
	}

	if (need_resetup)
		resetup();
}
