#include <unistd.h>
#include <malloc.h>
#include <limits.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/wait.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>

#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <netdb.h>
#include <syslog.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <string.h>

#include "inetd.h"

#define	TOOMANY		40		/* don't start more than TOOMANY */
#define	CNT_INTVL	60		/* servers in CNT_INTVL sec. */
#define	RETRYTIME	(60*5)		/* retry after bind or server fail */
#define RECYCLETIME	(60*15)

#ifdef DEBUG
int	debug = 0;
#endif

static int	nsock, maxsock, max_inetdsock;
fd_set	allsock;

static char * config_name = "/etc/inetd.conf";

#ifdef FAKE_SOMEBODY
static char ** Argv;
static char *  LastArg;
#endif

static int run_server(struct servtab*, struct port*, fd_set*);

static char need_exit = 0;
static char need_terminate = 0;
static char need_resetup = 0;
static char need_reconfig = 0;
static char need_reap_childs = 0;

static void reap_childs(void);
static void closedown(void) __attribute__ ((noreturn));
static void closedown_with_childs(void) __attribute__ ((noreturn));
static void resetup(void);
static void reconfig(void);

static void dolog(struct port *, int);
static char *rfc931_name(struct sockaddr_in * there, int ctrl);

static const char INETDPID[] = "/var/run/inetd.pid";
struct servtab * servtab;

void
logpid() {
	FILE *fp;

	if ((fp = fopen(INETDPID, "r")) != NULL) {
	        unsigned pid;
		if (fscanf(fp, "%u", &pid) == 1 && kill(0, pid)==0) {
	          logerr("inetd already runs!");
		  exit(-1);
	        }   
		(void)fclose(fp);
	}

	if ((fp = fopen(INETDPID, "w")) != NULL) {
		fprintf(fp, "%u\n", getpid());
		(void)fclose(fp);
	}
}

void
cleanup(int code) {
  unlink(INETDPID);
  exit(code);
}


/* Signal handlers */

static 
void
sigsegv_handler(int sig) {
  struct timeval tv;
  int i;
  tv.tv_sec = 10;
  tv.tv_usec = 0;
  logerr("signal %d got", sig);
  for (i=0; i < OPEN_MAX; i++)
    close(i);
  select(0,NULL,NULL,NULL,&tv);
  sigsetmask(0);
  execl("/sbin/inetd", "inetd", "-d", "2", NULL);
  cleanup(-1);
}


static void
sigdebug_handler(int sig) {
  debug++;
  if (sig == SIGUSR2)
    debug = 0;
  syslog(LOG_DEBUG, "debuggind level set to %d", debug);
}

static void
sigint_handler(int sig) {
  need_exit = 1;
}

static void
sigterm_handler(int sig) {
  need_terminate = 1;
}


static 
void
sigalrm_handler(int sig) {
  need_resetup = 1;
}

static void
sighup_handler(int sig) {
  need_reconfig = 1;
}

static void
sigchld_handler(int sig) {
  need_reap_childs = 1;
}

static char timingout=0;

static void
set_timeout(void) {
  if (!timingout) {
    timingout = 1;
    alarm(RETRYTIME);
  }
}

__inline__
static
void
reset_timeout(void) {
  timingout = 0;
  alarm(RECYCLETIME);
}

void
check_descriptors(char * desc, int n, fd_set * fds) {
	int i;
	struct servtab * sep;
	static int errcnt = 0;

	if (++errcnt > 2) {
		logerr("panic!\n");
		for (sep = servtab; sep; sep = sep->next) {
			if (sep->wait && sep->pid)
				kill(SIGKILL, sep->pid); 
		}
		sigsegv_handler(0);
		/* DOES NOT RETURN */
		exit(1);
	}
		
	for (i=0; i<n; i++) {
		struct timeval tv;
		fd_set tfds;
		int res;

		tv.tv_sec = tv.tv_usec = 0;
		FD_ZERO(&tfds);
		FD_SET(i, &tfds);
		res = select(i+1, &tfds, NULL, NULL, &tv);
		if (res < 0 && errno == EBADF) {
			struct port * port;
			logerr("found desc %d (%m) scanning %s\n", i, desc);
			FD_CLR(i, &fds);
			for (sep = servtab; sep; sep = sep->next) {
				for (port = sep->port; port; port = port->next) {
					if (port->fd != i)
						continue;
					logerr("svc %s\n", sep->server);
					if (sep->wait && sep->pid)
						kill(SIGKILL, sep->pid);
					port->fd = -1;
					sep->pid = 0;
					sep->obsolete = 1;
					debug = 1;
					need_resetup = 1;
				}
			}
		}
	}
}

void
main (int argc, char * argv[], char *envp[]) {
#ifdef FAKE_SOMEBODY
  Argv = argv;
  if (envp == 0 || *envp == 0)
    envp = argv;
  while (*envp)
    envp++;
  LastArg = envp[-1] + strlen(envp[-1]);
#endif

#ifdef DEBUG
  _BLOCK_BEGIN
  int ch;

  while ((ch = getopt(argc, argv, "d:")) != EOF) {
    switch(ch) {
    case 'd':
      debug = atoi(optarg);
      break;
    case '?':
    default:
      fprintf(stderr, "usage: inetd [-d level]\n");
      exit(1);
    }
  }
  argc -= optind;
  argv += optind;
  _BLOCK_END
#endif

  if (argc > 0) {
    config_name = argv[0];
    if (debug)
      fprintf(stderr, "config is \"%s\"\n", config_name);
  }

  _BLOCK_BEGIN
  extern void daemon(int,int);
  daemon(0, 0);

  _BLOCK_END
  openlog("inetd", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
  logpid();
   
  _BLOCK_BEGIN
  struct sigvec sv;
  bzero((char *)&sv, sizeof(sv));
  sv.sv_mask = 0;
  sv.sv_handler = sigalrm_handler;
  sigvec(SIGALRM, &sv, (struct sigvec *)0);
  sv.sv_handler = sighup_handler;
  sigvec(SIGHUP, &sv, (struct sigvec *)0);
  sv.sv_handler = sigchld_handler;
  sigvec(SIGCHLD, &sv, (struct sigvec *)0);
  sv.sv_handler = sigterm_handler;
  sigvec(SIGTERM, &sv, (struct sigvec *)0);
  sv.sv_handler = sigint_handler;
  sigvec(SIGINT, &sv, (struct sigvec *)0);
  sv.sv_handler = sigdebug_handler;
  sigvec(SIGUSR1, &sv, (struct sigvec *)0);
  sigvec(SIGUSR2, &sv, (struct sigvec *)0);
  sv.sv_handler = sigsegv_handler;
  sigvec(SIGSEGV, &sv, (struct sigvec *)0);
  need_reconfig = 1;
  _BLOCK_END

#ifdef FAKE_SOMEBODY
  _BLOCK_BEGIN
  /* space for daemons to overwrite environment for ps */
# define	DUMMYSIZE	100
  char dummy[DUMMYSIZE];
  memset(dummy, 'x', sizeof(dummy) - 1);
  dummy[DUMMYSIZE - 1] = '\0';
  setenv("inetd_dummy", dummy, 1);
# undef DUMMYSIZE
  _BLOCK_END
#endif

  for (;;) {
    if (need_exit)
      closedown();
    if (need_terminate)
      closedown_with_childs();
    if (need_reap_childs) {
      reap_childs();
      continue;
    }
    if (need_reconfig) {
      reconfig();
      continue;
    }
    if (need_resetup) {
      resetup();
      continue;
    }
    if (nsock == 0) {
      pause();
      continue;
    }

    _BLOCK_BEGIN
    int n;
    fd_set readable;
    extern fd_set svc_fdset;
    extern void svc_getreqset(fd_set *);
    
    readable = allsock;
    maxsock = FDSET_OR(&readable, &svc_fdset);
    maxsock = (maxsock > max_inetdsock ? maxsock : max_inetdsock);

    if ((n = select(maxsock + 1, &readable, NULL, NULL, NULL)) <= 0) {
      if (n < 0 && errno != EINTR) {
	if (errno == EBADF) {
		logerr("select: %m\n");
		check_descriptors("allsock", maxsock+1, &allsock);
		check_descriptors("allsock", maxsock+1, &svc_fdset);
	}
  	logerr("select: %m\n");
  	sleep(1);
      }
      continue;
    }

    _BLOCK_BEGIN
    register struct servtab *sep;
    register struct port *port;
    int i;
    int rpc_n=0;
    fd_set rpc_readable;

    FD_ZERO(&rpc_readable); 
    for (i=0; i<OPEN_MAX; i++) {
      if (FD_ISSET(i,&readable)) {
	if (FD_ISSET(i,&svc_fdset)) {
	  FD_SET(i,&rpc_readable);
	  FD_CLR(i,&readable);
	  n--;
	  rpc_n++;
        }
      }
    }
    if (rpc_n) {
/* RPC library is so buggy, that EINTR may be fatal for it */
      int saved_sigmask = siggetmask();
      sigsetmask(saved_sigmask|sigmask(SIGALRM)|sigmask(SIGUSR1)|
		               sigmask(SIGUSR2)|sigmask(SIGHUP)|
		               sigmask(SIGCHLD)
		);
      svc_getreqset(&rpc_readable);
      sigsetmask(saved_sigmask);
    }

    for (sep = servtab; n && sep; sep = sep->next) {
      for (port = sep->port; port; port = port->next) {
        if (port->fd != -1 && FD_ISSET(port->fd, &readable)) {
  	  dprintf(1, "someone wants %d", ntohs(IN_PORT(port)));
          n -= run_server(sep, port, &readable);
  	}
      }
    }
    _BLOCK_END
    _BLOCK_END
  }
}



/* Closedown procedure */

static void
closedown(void) {
  loginfo("inetd terminated");
  cleanup(0);
}

static void
closedown_with_childs(void) {
  struct servtab * sep;
  for (sep = servtab; sep; sep = sep->next) {
    if (sep->wait && sep->pid)
      kill(SIGTERM, sep->pid); 
  }
  loginfo("inetd terminated with childs");
  cleanup(0);
}

static struct port * leaved_ports = NULL;

void
activate_fd(int fd) {
  int  onoff = 0;
  if (fd < 0 || FD_ISSET(fd,&allsock))
    return;
  FD_SET(fd, &allsock);
  nsock++;
  if (fd > max_inetdsock)
    max_inetdsock = fd;
  ioctl(fd, FIONBIO, (char *) &onoff);
}

static
void
deactivate_fd(int fd) {
  if (fd < 0 || !FD_ISSET(fd,&allsock))
    return;
  FD_CLR(fd, &allsock);
  if (fd == max_inetdsock)
    max_inetdsock--;
  nsock--;
}


/* Port list tools */

static
void 
remove_from_port_list(struct port ** pport, struct port * port) {
#ifdef PARANOID
  if (!pport) {
    logerr("par1");
    cleanup(1);
  }
  if (!*pport) {
    logerr("par13");
    cleanup(1);
  }
  if ((*pport)->prev) {
    logerr("par21");
    cleanup(1);
  }
#endif
  if (port->next)
    port->next->prev = port->prev;
  if (port->prev)
    port->prev->next = port->next;
  if (*pport == port)
    *pport = port->next;
}

void 
add_to_port_list(struct port ** pport, struct port * port) {
#ifdef PARANOID
  if (!pport) {
    logerr("par2");
    cleanup(1);
  }
  if (*pport && (*pport)->prev) {
    logerr("par3");
    cleanup(1);
  }
#endif
  port->next = *pport;
  port->prev = NULL;
  if (*pport)
    (*pport)->prev = port;
  *pport = port;
}

static void 
free_port(struct port * port) {
  struct pmaplist * pml=port->pml;
  while (pml) {
    struct pmaplist * nextpml = pml->pml_next;
    free(pml);
    pml = nextpml;
  }
  free(port);
}

static void 
discard_port_list(struct port ** pport) {
  register struct port * port = *pport;
  register struct port * next;
#ifdef PARANOID
  if (!pport) {
    logerr("par4");
    cleanup(1);
  }
#endif
  while (port) {
    next = port->next;
    remove_from_port_list(pport, port);
    if (port->fd < 0 || (port->pml && port->pml->pml_map.pm_prog == PMAPPROG))
      free_port(port);
    else
      add_to_port_list(&leaved_ports, port);
    port = next;
  }  
}

void 
free_port_list(struct port ** pport) {
  register struct port * port = *pport;
  register struct port * next;
#ifdef PARANOID
  if (!pport) {
    logerr("par5");
    cleanup(1);
  }
#endif
  while (port) {
    next = port->next;
    remove_from_port_list(pport, port);
    if (port->fd >= 0) {
      deactivate_fd(port->fd);
      close(port->fd);
    }
    free_port(port);
    port = next;
  }  
}

__inline__ static int
ports_match(struct port * port1, struct port * port2) {
  return 
#ifdef LISTEN_OTHER_FAMILIES
         FAMILY(port1) == FAMILY(port2) &&
         IN_ADDR(port1) == IN_ADDR(port2) &&
#endif
         IN_PORT(port1) == IN_PORT(port2) &&
         port1->socktype == port2->socktype &&
         port1->proto == port2->proto;
}

static struct servtab * 
lookup_port(struct port * port) {
  struct servtab  * sep;
  struct port  * p;
  for (sep = servtab; sep; sep = sep->next) {
    for (p = sep->port; p; p = p->next) {
      if (ports_match(port,p))
	return sep;
    } 
  }
  return NULL;
}

static struct port *
find_port_in_list(struct port * port, struct port * list) {
  register struct port  * p;
  for (p = list; p; p = p->next) {
    if (ports_match(port,p))
      return p;
  }
  return NULL;
}



/* servtab list tools */

static void 
free_servtab(struct servtab *cp) {
  int i;

  free_port_list(&cp->port);
  if (cp->user)
    free(cp->user);
  if (cp->server)
    free(cp->server);
  for (i = 0; i < MAXARGV; i++)
    if (cp->argv[i])
      free(cp->argv[i]);
}

static void 
remove_from_servtab_list(struct servtab ** sepp, struct servtab * sep) {
#ifdef PARANOID
  if (!sepp) {
    logerr("par6");
    cleanup(1);
  }
  if (!*sepp) {
    logerr("par12");
    cleanup(1);
  }
  if ((*sepp)->prev) {
    logerr("par20");
    cleanup(1);
  }
#endif
  if (sep->next) 
    sep->next->prev = sep->prev;
  if (sep->prev)
    sep->prev->next = sep->next;
  if (*sepp == sep)
    *sepp = sep->next;
}


static void 
add_to_servtab_list(struct servtab ** sepp, struct servtab * sep) {
#ifdef PARANOID
  if (!sepp) {
    logerr("par7");
    cleanup(1);
  }
  if (*sepp && (*sepp)->prev) {
    logerr("par8");
    cleanup(1);
  }
  if (!sep) {
    logerr("par9");
    cleanup(1);
  }
#endif
  sep->next = *sepp;
  sep->prev = NULL;
  if (*sepp)
    (*sepp)->prev = sep;
  *sepp = sep;
}


void 
discard_servtab(struct servtab * sep) {
  discard_port_list(&sep->port);
  free_servtab(sep);
}

static void 
discard_servtab_list(struct servtab ** sepp) {
  struct servtab *sep, *next;

#ifdef PARANOID
  if (!sepp) {
    logerr("par10");
    cleanup(1);
  }
#endif
  sep = *sepp;
  while (sep) {
    next = sep->next;
    if (!sep->wait || sep->pid == 0) {
      remove_from_servtab_list(sepp, sep);
      discard_servtab(sep);
    }
    sep = next;
  }
}


static void 
reconfig(void) {
  register struct servtab *sep;
  struct port *port;
  time_t curtime;

  time(&curtime);

  dprintf(1,"reconfig");

  if (!setconfig(config_name)) {
    logerr( "cannot open %s: %m", config_name);
    if (!servtab)
      cleanup(1);
    logerr( "reconfig command ignored");
    need_reconfig = 0;
    return;
  }

  setrpcent(1);

  discard_servtab_list(&servtab);

  for (sep = servtab; sep; sep = sep->next) {
    sep->obsolete = 1; 
    dprintf(1, "%s still running", sep->server);
  }

  shutdown_portmap();

more:

  while ((sep = getconfigent()) != NULL) {
    for (port = sep->port; port; port=port->next) {
      struct servtab * s;

      port->last_use = curtime;
      if (!IN_PORT(port) && !port->pml) {
	logerr( "port for %s is not resolved, ignored.", sep->server);
	discard_servtab(sep);
	goto more;
      }
      if (IN_PORT(port)) {
        if ((s = lookup_port(port)) != NULL) {
          if (!s->obsolete) {
	    logerr("port %d listened by >1 services", ntohs(IN_PORT(port)));
	    discard_servtab(sep);
	    goto more;
	  }
	  port->fd = -2;
        } else if (find_port_in_list(port,leaved_ports))
	  port->fd = -2;
      } else {
	struct port     * oldport = NULL;
	struct pmaplist * pml;

	for (pml = port->pml; pml; pml = pml->pml_next) {
	  struct port     * tport;
	  if (find_rpc_service(pml->pml_map.pm_prog,
			       pml->pml_map.pm_vers,
			       pml->pml_map.pm_prot, &s, &tport)) {
            if (!s->obsolete) {
	      logerr( "rpc%d listened by >1 services", pml->pml_map.pm_prog);
	      discard_servtab(sep);
	      goto more;
	    }
	    if (IN_PORT(tport))
	      oldport = tport;
	  } else
	      dprintf(1, "rpc%d/%d/%d not found\n", pml->pml_map.pm_prog,
		      pml->pml_map.pm_vers,
		      pml->pml_map.pm_prot);
	}
	if (oldport) {
	  port->fd = -2;
	  IN_PORT(port) = IN_PORT(oldport);
          for (pml = port->pml; pml; pml = pml->pml_next) {
		  pml->pml_map.pm_port = ntohs(IN_PORT(oldport));
		  dprintf(1, "rpc%d/%d/%d waiting for port %d\n",
			  pml->pml_map.pm_prog,
			  pml->pml_map.pm_vers,
			  pml->pml_map.pm_prot,
			  pml->pml_map.pm_port);
	  }
	}
      }
    } 
    add_to_servtab_list(&servtab, sep);
  }
  endrpcent();
  endconfig();
  need_reconfig = 0;
  resetup();
}


void
set_port_options(struct servtab * sep, struct port * port) {
  int off_on[2] = {0, 1};

#ifdef DEBUG
  if (setsockopt(port->fd, SOL_SOCKET, SO_DEBUG, (char *)&off_on[debug > 1 || port->debug], 
#else
  if (setsockopt(port->fd, SOL_SOCKET, SO_DEBUG, (char *)&off_on[port->debug], 
#endif
                                                sizeof (int))<0) {
      logerr( "setsockopt (SO_DEBUG): %m");
  }

  if (setsockopt(port->fd, SOL_SOCKET, SO_REUSEADDR, 
                 (char *)&off_on[port->reuse], sizeof (int))<0) {
      logerr( "setsockopt (SO_REUSEADDR) = %i: %m", port->reuse);
  }
  if (setsockopt(port->fd, SOL_SOCKET, SO_BROADCAST, 
                 (char *)&off_on[port->broadcast], sizeof (int))<0) {
      logerr( "setsockopt (SO_BROADCAST) = %i: %m", port->broadcast);
  }
  if (setsockopt(port->fd, SOL_SOCKET, SO_KEEPALIVE, 
                 (char *)&off_on[port->keepalive], sizeof (int))<0) {
      logerr( "setsockopt (SO_KEEPALIVE) = %i: %m", port->keepalive);
  }
}


static void
flush_port(struct port * port, fd_set * readable) {
  if (port->fd < 0)
    return;
  if (FD_ISSET(port->fd, readable)) {
    char buf[50];
    FD_CLR(port->fd, readable);
    if (port->socktype != SOCK_STREAM)
      recv(port->fd, buf, sizeof (buf), 0);
  }
}

static void
flush_dgram_ports(fd_set * readable) {
  int n;
  for (n = 0; n <= max_inetdsock; n++)
    if (FD_ISSET(n, readable)) {
      char buf[50];
      FD_CLR(n, readable);
      recv(n, buf, sizeof (buf), 0);
    }
}

static int
count_ports(struct servtab * sep, fd_set *readable) {
  struct port * p;
  int nr = 0;
#ifdef PARANOID
  if (!sep->wait) {
    logerr("par15");
    cleanup(1);
  }
#endif  
  for (p = sep->port; p; p = p->next) {
#ifdef PARANOID
    if (p->fd < 0) {
      logerr("par16");
      cleanup(1);
    }
#endif
    if (FD_ISSET(p->fd, readable)) 
      nr++;
  }
#ifdef PARANOID
  if (!nr) {
    logerr("par17");
    cleanup(1);
  }
#endif
  return nr;
}

static void
deactivate_service(struct servtab * sep) {
  struct port * port;
  for (port = sep->port; port; port = port->next) 
    deactivate_fd(port->fd);
}

static int
activate_service(struct servtab * sep) {
  struct port * port;
  time_t curtime;

  time(&curtime);

  for (port = sep->port; port; port = port->next) {
    port->last_use =  curtime;
    if (port->fd < 0) {
      deactivate_service(sep);
      return -1;
    }
  }
  if (sep->port->pml && sep->port->pml->pml_map.pm_prog == PMAPPROG)
    return resetup_portmap(sep);
  for (port = sep->port; port; port = port->next) 
    activate_fd(port->fd);
  return 0;
}


static int
open_port(struct servtab * sep, struct port * port) {
  int fd;
  if ((fd = socket(FAMILY(port), port->socktype, 0)) < 0) {
    logerr( "socket: %i: %m", ntohs(IN_PORT(port)));
    return -1;
  }
  if (!sep->wait) {
    int on = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, 
                 (char *)&on, sizeof (on))<0)
      logerr( "open_port: setsockopt (SO_REUSEADDR) = 1: %m");
  }
  if (bind(fd, &port->addr, sizeof (port->addr)) < 0) {
    logerr( "bind: %i: %m", ntohs(IN_PORT(port)));
    close(fd);
    set_timeout(); 
    return -1;
  }
  port->fd = fd;
  set_port_options(sep, port);
  if (port->socktype == SOCK_STREAM)
    listen(fd, 10);
  activate_fd(fd);
  return fd;
}


int
bind_rpc_service(struct servtab * sep) {
  struct port * port;
  struct pmaplist * pml;
  time_t curtime;
  time(&curtime);

  for (port = sep->port; port; port = port->next) {
    int fd;
    u_long prog, prot;
    port->last_use = curtime;

    if (IN_PORT(port))
      continue;
#ifdef PARANOID
    if (!port->pml) {
      logerr("par100");
      cleanup(1);
    }
#endif
    prog = port->pml->pml_map.pm_prog; 
    prot = port->pml->pml_map.pm_prot; 
    if ((fd = socket(FAMILY(port), port->socktype, 0)) < 0) {
      logerr( "socket: rpc%d/%d: %m", prog, prot);
      return 0;
    }
    if (bindresvport(fd, (struct sockaddr_in*)&port->addr) < 0) {
      logerr( "bindresvport: rpc%d/%d: %m", prog, prot);
      close(fd);
      return 0;
    }
    port->fd = fd;
    set_port_options(sep, port);
    for (pml = port->pml; pml; pml=pml->pml_next) {
      dprintf(2, "rpc%d bound to %d at %d", pml->pml_map.pm_prog,
	     ntohs(IN_PORT(port)), fd);
      pml->pml_map.pm_port = ntohs(IN_PORT(port));
    }
    if (port->socktype == SOCK_STREAM)
      listen(fd, 5);
    activate_fd(fd);
  }
  return 1;
}


static void
resetup() {
  register struct servtab * sep;
  register struct port * port;
  time_t curtime;
  
  dprintf(1,"resetup");
  time(&curtime);

  reset_timeout();

  for (sep = servtab; sep; sep = sep->next) {
    if (sep->obsolete || (sep->wait && sep->pid))
      continue;
    for (port = sep->port; port; port = port->next) {
      if (port->autorpc && IN_PORT(port) && 
	  curtime >= port->last_use + RECYCLETIME*2) {
        struct pmaplist * pml;
#ifdef PARANOID
	if (!port->pml) {
	  logerr("par300");
	  cleanup(1);
	}
#endif
	dprintf(1, "port%i/%i is recycled\n", ntohs(IN_PORT(port)), port->proto);
	for (pml = port->pml; pml; pml = pml->pml_next)
	  pml->pml_map.pm_port = 0;
        if (port->fd >= 0) {
          deactivate_fd(port->fd);
          close(port->fd);
	  port->fd = -1;
        }
	IN_PORT(port) = 0;
      }
    }
  }


  for (sep = servtab; sep; sep = sep->next) {
    int something_done = 0;
    if (sep->obsolete || (sep->wait && sep->pid))
      continue;
    for (port = sep->port; port; port = port->next) {
      switch (-port->fd) {
      case 1:
	if (IN_PORT(port)) {
	  if (open_port(sep, port) >= 0)
            something_done = 1;
          dprintf(1, "opening %i -> %i", ntohs(IN_PORT(port)), port->fd);
        }
	continue;
      case 2:
	{ 
	  struct port * p;
          if ((p = find_port_in_list(port, leaved_ports))  !=  NULL) {
            dprintf(1, "restoring %i -> %i", ntohs(IN_PORT(port)), p->fd);
	    remove_from_port_list(&leaved_ports, p);
	    port->fd = p->fd;
	    p->fd = -1;
            something_done = 1;
	    free_port(p);
	    set_port_options(sep, port);
          } else
            dprintf(1, "restore %i failed", ntohs(IN_PORT(port)));
	  continue;
        }
      default:
	continue;
      } 
    } 
    if (something_done)
      activate_service(sep);
  }
#ifdef DEBUG
  for (port = leaved_ports; port; port = port->next)
    dprintf(1, "port %i is cancelled", ntohs(IN_PORT(port)));
#endif
  free_port_list(&leaved_ports);
  need_resetup = 0;
}


static pid_t 
fork_child(struct servtab *sep, struct port * port, fd_set * readable) {
  if (sep->count++ == 0)
    gettimeofday(&sep->start_time, NULL);
  else if (sep->count >= TOOMANY) {
    struct timeval now;
    gettimeofday(&now, NULL);
    if (now.tv_sec - sep->start_time.tv_sec > CNT_INTVL) {
      sep->start_time = now;
      sep->count = 1;
    } else {
      logerr( "%s server, service terminated", sep->server);
      sep->count = 0;

      set_timeout();
      errno = 0;
      return -1;
    }
  }
  return fork();
}

/* These two procedures started only after forking,
 * they rearrange file descriptors to pass them to a service.  
 */

static void
fork_accepted_port(struct servtab * sep, int ctrl) {
  int i;
  dprintf(1, "forked one port");
#ifdef PARANOID
  if (ctrl <= 2) {
    logerr("fork_accepted_port : ctrl<=2");
    return; 
  }  
#endif
  for (i = OPEN_MAX-1; i > 2; i--)
    if (i != ctrl)
      close(i);
  dup2(ctrl, 0);
  close(ctrl);
  dup2(0, 1);
  dup2(0, 2);
}

static void
fork_multi_ports(struct servtab * sep, fd_set * readable) {
  int n,m;
  dprintf(1, "forked multi port");
  FD_ZERO(readable);
  for (n=0, m=0; n <= max_inetdsock; n++) {
    if (FD_ISSET(n, &allsock)) {
      struct port * p;
      for (p = sep->port; p; p = p->next) {
      	if (p->fd == n) 
      	  break;
      }
      if (!p)
      	continue;
      dprintf(1, "%i/%i socket(%i) dupped to %i", ntohs(IN_PORT(p)), 
	      p->proto, n, m);
      if (n != m)
      	dup2(n, m);
      if (p->socktype != SOCK_STREAM)
        FD_SET(m, readable);
      m++;
    }
  }

  if (m < 3) {
    int tfd = open("/dev/null",O_RDWR);
    for (; m < 3 ; m++) {
      if (m != tfd)
        dup2(tfd, m);
    } 
    close(tfd);
  }
  for (; m < OPEN_MAX; m++)
    close(m);
} 


static int
run_server(struct servtab *sep, struct port *port, fd_set *readable) {
  struct port * p;
  int ctrl = -1;

  if (sep->wait) { 
    for (p = sep->port; p; p = p->next) {
      if (p->fd < 0) {
        logerr("cannot start %s, port is busy", sep->server);
        deactivate_service(sep);
        for (p = sep->port; p; p = p->next) 
	  flush_port(p, readable);
        return count_ports(sep, readable);
      }
    }
  }

  if (!sep->wait) {
    int onoff = 1;
    switch (port->socktype) {
    case SOCK_STREAM:
      ioctl(port->fd, FIONBIO, (char *) &onoff);
      ctrl = accept(port->fd, NULL, NULL);
      onoff = 0;
      ioctl(port->fd, FIONBIO, (char *) &onoff);
      break;
    default:
      ctrl = dup(port->fd);
    }
    if (ctrl < 0) {
      if (errno !=  EINTR && errno != EWOULDBLOCK && errno != EAGAIN)
        logerr( "accept (for %i): %m", ntohs(IN_PORT(port)));
      return 1;
    }
    onoff = 1;
    if (setsockopt(ctrl, SOL_SOCKET, SO_REUSEADDR, 
                 (char *)&onoff, sizeof (onoff))<0) {
      logerr("setsockopt (SO_REUSEADDR) of accepted port: %m");
    }
    dprintf(1, "accepted %i", ntohs(IN_PORT(port)));
  }

  /* No fork needed */

  if (sep->bi && !sep->bi->forking) {
    (*sep->bi->bi_fn)( (ctrl >=0 ? ctrl : port->fd ), sep);
    if (ctrl >= 0)
      close(ctrl);
    return 1;
  };


  _BLOCK_BEGIN
  int pid;
  pid = fork_child(sep, port, readable); 

  if (pid < 0) {                                /* Fork failure */
    if (errno) {
      logerr( "fork: %m");
      sleep(1);
    } 

    if (sep->wait) {
      struct port * p;
      int nr = count_ports(sep, readable);
      deactivate_service(sep);
      for (p = sep->port; p; p = p->next) {
	flush_port(p, readable);
        close(p->fd);
	p->fd = -1;
      }
      return nr;
    } else {
      deactivate_fd(port->fd);
      flush_port(port, readable);
      close(port->fd);
      port->fd = -1;
      if (ctrl >= 0)
        close(ctrl);
      return 1;
    }
  } else if (pid) {                        /* Fork done; parent side */
    if ( sep->wait) {
      struct port * p;
      int nr = 0;
      sep->pid = pid;
      deactivate_service(sep);
      for (p = sep->port; p; p=p->next) {
        if (p->fd >= 0) {
          FD_CLR(p->fd, readable);
	  nr++;
        }
      }
      return nr;
    } else {
      if (ctrl >= 0)
        close(ctrl);
      return 1;
    }
  } else {                                      /* Fork done; child side */
    openlog("inetd_child", LOG_PID, LOG_DAEMON);

    if (port->log)
      dolog(port, ctrl >=0 ? ctrl : port->fd);

    if (ctrl >= 0) 
      fork_accepted_port(sep, ctrl);
    else
      fork_multi_ports(sep, readable);

    if (sep->bi) {
      (*sep->bi->bi_fn)( 0, sep);
      _exit(0);
    } else {
      register struct passwd *pwd;
      dprintf(1, "%d execl %s\n", getpid(), sep->server);
      if ((pwd = getpwnam(sep->user)) == NULL) {
        logerr( "getpwnam: %s: No such user", sep->user);
        flush_dgram_ports(readable);         
        _exit(-1);
      }
      if (pwd->pw_uid) {
        setgid((gid_t)pwd->pw_gid);
        initgroups(pwd->pw_name, pwd->pw_gid);
        setuid((uid_t)pwd->pw_uid);
      }
      execv(sep->server, sep->argv);
      logerr( "execv %s: %m", sep->server);
      flush_dgram_ports(readable);         
      _exit(1);
    }
    /*NOT REACHED*/  
  }
  _BLOCK_END
  /*NOT REACHED*/  
}


/* Procedure handling reaped childs */

static void
reap_childs(void) {
  char * name = "rpc_callit";

  for (;;) {
    int status;
    int pid;
    register struct servtab * sep;
    pid = wait3(&status, WNOHANG, NULL);
    if (pid <= 0) {
      need_reap_childs = 0;
      break;
    }
    sep = servtab;
    for (sep = servtab; sep; sep = sep->next) {
      if (sep->wait && sep->pid == pid) {
	name = sep->server;
        break;
      }
    }
    if (!sep && !reap_rpc_callit(pid))
      continue;
    dprintf(1, "%s[%i] reaped", name, pid);
    if (WIFEXITED(status)) {
      if (WEXITSTATUS(status)) {
 	logerr("%s: exited with status %i",
 		            name, WEXITSTATUS(status));
      }
    } else if (WIFSIGNALED(status)) {
      logerr("%s: got signal \"%s\"",
 		          name, sys_siglist[WTERMSIG(status)]);
    } else if (WIFSTOPPED(status)) {
      logerr("%s: stopped by signal \"%s\"",
 		          name, sys_siglist[WSTOPSIG(status)]);
    } else {
      logerr( "%s: exited with invalid status 0x%08x",
 		      name, status);
    }
 		
    if (!sep)
      continue;

    if (sep->obsolete) {
      discard_servtab(sep);
      need_resetup = 1;
    } else {
      sep->pid = 0;
      activate_service(sep);
    }
  }

  if (need_resetup)
    resetup();
}

static void 
dolog(struct port * port, int ctrl) {
	struct sockaddr		sa;
	struct sockaddr_in	*sin = (struct sockaddr_in *)&sa;
	int			len = sizeof(sa);
	struct hostent		*hp;
	const char		*host;
	char			buf[BUFSIZ];
	int			connected = 1;
	static char             * curdom = NULL;

        if (!curdom) {
	  char dbuf[MAXHOSTNAMELEN]; 
          if (getdomainname(dbuf, sizeof(dbuf)) >= 0) {
	    curdom = malloc(strlen(dbuf)+1);
	    if (curdom) 
	      strcpy(curdom, dbuf);
          }
        }

	if (getpeername(ctrl, &sa, &len) < 0) {
		if (errno != ENOTCONN) {
			logerr( "getpeername: %m");
			return;
		}
		if (recvfrom(ctrl, buf, sizeof(buf), MSG_PEEK, &sa, &len) < 0) {
			logerr( "recvfrom: %m");
			return;
		}
		connected = 0;
	}
	if (sa.sa_family != AF_INET) {
		logerr( "unexpected address family %u", sa.sa_family);
		return;
	}

	hp = gethostbyaddr((char *) &sin->sin_addr.s_addr,
				sizeof (sin->sin_addr.s_addr), AF_INET);

	host = hp ? hp->h_name : inet_ntoa(sin->sin_addr);

        if (strchr(host, '.') == 0)
	  return;

	if (curdom && port->rfc931) {
	        const char * dp;
		dp = host + strlen(host) - strlen(curdom);
		if (dp >= host && strcasecmp(dp, curdom) == 0)
		  return;
	}

	if (connected && port->rfc931)
		syslog(LOG_AUTH, "%s@%s wants %i",
				rfc931_name(sin, ctrl), host, ntohs(IN_PORT(port)));
	else
		syslog(LOG_AUTH, "%s wants %i",	host, ntohs(IN_PORT(port)));
}

/*
 * From tcp_log by
 *  Wietse Venema, Eindhoven University of Technology, The Netherlands.
 */

#include <setjmp.h>

#define	RFC931_PORT	113		/* Semi-well-known port */
/*
#define	TIMEOUT		4
#define	TIMEOUT2	10
 */
#define	TIMEOUT		10

static jmp_buf timebuf;

/* timeout - handle timeouts */

static void timeout(int sig) {
	longjmp(timebuf, sig);
}

/* rfc931_name - return remote user name */

static char *
rfc931_name(struct sockaddr_in * there, int ctrl) {
	struct sockaddr_in here;	/* local link information */
	struct sockaddr_in sin;		/* for talking to RFC931 daemon */
	int		length;
	int		s;
	unsigned	remote;
	unsigned	local;
	static char	user[256];		/* XXX */
	char		buf[256];
	char		*cp;
	static char	*result = "USER_UNKNOWN";
	int		len;

	/* Find out local port number of our stdin. */

	length = sizeof(here);
	if (getsockname(ctrl, (struct sockaddr *) &here, &length) == -1) {
		logerr( "getsockname: %m");
		return (result);
	}
	/* Set up timer so we won't get stuck. */

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		logerr( "socket: %m");
		return (result);
	}

	sin = here;
	sin.sin_port = htons(0);
	if (bind(s, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
		logerr( "bind: %m");
		return (result);
	}

	signal(SIGALRM, timeout);
	if (setjmp(timebuf)) {
		close(s);			/* not: fclose(fp) */
		return (result);
	}
	alarm(TIMEOUT);

	/* Connect to the RFC931 daemon. */

	sin = *there;
	sin.sin_port = htons(RFC931_PORT);
	if (connect(s, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
		close(s);
		alarm(0);
		return (result);
	}

	/* Query the RFC 931 server. Would 13-byte writes ever be broken up? */
	sprintf(buf, "%u,%u\r\n", ntohs(there->sin_port), ntohs(here.sin_port));


	for (len = 0, cp = buf; len < strlen(buf); ) {
		int	n;
		if ((n = write(s, cp, strlen(buf) - len)) == -1) {
			close(s);
			alarm(0);
			return (result);
		}
		cp += n;
		len += n;
	}

	/* Read response */
	for (cp = buf; cp < buf + sizeof(buf) - 1; ) {
		char	c;
		if (read(s, &c, 1) != 1) {
			close(s);
			alarm(0);
			return (result);
		}
		if (c == '\n')
			break;
		*cp++ = c;
	}
	*cp = '\0';

	if (sscanf(buf, "%u , %u : USERID :%*[^:]:%255s", &remote, &local, user) == 3
		&& ntohs(there->sin_port) == remote
		&& ntohs(here.sin_port) == local) {

		/* Strip trailing carriage return. */
		if ((cp = strchr(user, '\r')) != NULL)
			*cp = 0;
		result = user;
	}

	alarm(0);
	close(s);
	return (result);
}
