/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
/*
 * Copyright (c) 1986-1991 by Sun Microsystems Inc. 
 */

#pragma ident	"@(#)svcauth_des.c	1.18	94/04/24 SMI"

/*
 * svcauth_des.c, server-side des authentication
 *
 * We insure for the service the following:
 * (1) The timestamp microseconds do not exceed 1 million.
 * (2) The timestamp plus the window is less than the current time.
 * (3) The timestamp is not less than the one previously
 *	seen in the current session.
 *
 * It is up to the server to determine if the window size is
 * too small.
 *
 */

#include <assert.h>
#include "rpc_mt.h"
#include <rpc/des_crypt.h>
#include <rpc/rpc.h>
#include <sys/types.h>
#include <rpc/trace.h>
#include <stdlib.h>
#include <string.h>

#ifdef KERNEL
#include <sys/kernel.h>
#define	gettimeofday(tvp, tzp)	(*(tvp) = time)
#else
#include <sys/syslog.h>
#endif

#define	USEC_PER_SEC	1000000L
#define	BEFORE(t1, t2) timercmp(t1, t2, < /* EMPTY */)

/*
 * LRU cache of conversation keys and some other useful items.
 */
#define	AUTHDES_CACHESZ 64
struct cache_entry {
	des_block key;			/* conversation key */
	char *rname;			/* client's name */
	u_int window;			/* credential lifetime window */
	struct timeval laststamp;	/* detect replays of creds */
	char *localcred;		/* generic local credential */
};
struct cache_entry *_rpc_authdes_cache;	/* [AUTHDES_CACHESZ] */
static short *authdes_lru;		/* [AUTHDES_CACHESZ] */
/*
 *	A rwlock_t would seem to make more sense, but it turns out we always
 *	muck with the cache entries, so would always need a write lock (in
 *	which case, we might as well use a mutex).
 */
extern mutex_t	authdes_lock;


static void cache_init(void);		/* initialize the cache */
static short cache_victim(void);
static short cache_spot(des_block *key, char *name, struct timeval *timestamp);
                                        /* find an entry in the cache */
static void cache_ref(short sid);	/* note that sid was ref'd */
static void __msgout(const char *, const char *);

/*
 * cache statistics
 */
struct {
	u_long ncachehits;	/* times cache hit, and is not replay */
	u_long ncachereplays;	/* times cache hit, and is replay */
	u_long ncachemisses;	/* times cache missed */
} svcauthdes_stats;

/*
 * Service side authenticator for AUTH_DES
 */
enum auth_stat
__svcauth_des(rqst, msg)
	register struct svc_req *rqst;
	register struct rpc_msg *msg;
{

	register __u32	*ixdr;
	des_block	cryptbuf[2];
	register struct authdes_cred	*cred;
	struct authdes_verf	verf;
	int	status;
	register struct cache_entry	*entry;
	short	sid;
	des_block	*sessionkey;
	des_block	ivec;
	u_int	window;
	struct timeval	timestamp;
	u_long	namelen;
	struct area {
		struct authdes_cred area_cred;
		char area_netname[MAXNETNAMELEN+1];
	} *area;
	int	fullname_rcvd = 0;

	trace1(TR___svcauth_des, 0);
	mutex_lock(&authdes_lock);
	if (_rpc_authdes_cache == NULL) {
		cache_init();
	}
	mutex_unlock(&authdes_lock);

	area = (struct area *)rqst->rq_clntcred;
	cred = (struct authdes_cred *)&area->area_cred;

	/*
	 * Get the credential
	 */
	ixdr = (__u32 *)msg->rm_call.cb_cred.oa_base;
	cred->adc_namekind = IXDR_GET_ENUM(ixdr, enum authdes_namekind);
	switch (cred->adc_namekind) {
	case ADN_FULLNAME:
		namelen = IXDR_GET_U_LONG(ixdr);
		if (namelen > MAXNETNAMELEN) {
			trace1(TR___svcauth_des, 1);
			return (AUTH_BADCRED);
		}
		cred->adc_fullname.name = area->area_netname;
		memcpy(cred->adc_fullname.name, (char *)ixdr, (u_int)namelen);
		cred->adc_fullname.name[namelen] = 0;
		ixdr += (RNDUP(namelen) / BYTES_PER_XDR_UNIT);
		cred->adc_fullname.key.key.high = *ixdr++;
		cred->adc_fullname.key.key.low = *ixdr++;
		cred->adc_fullname.window = *ixdr++;
		fullname_rcvd++;
		break;
	case ADN_NICKNAME:
		cred->adc_nickname = *ixdr++;
		break;
	default:
		trace1(TR___svcauth_des, 1);
		return (AUTH_BADCRED);
	}

	/*
	 * Get the verifier
	 */
	ixdr = (__u32 *)msg->rm_call.cb_verf.oa_base;
	verf.adv_xtimestamp.key.high = *ixdr++;
	verf.adv_xtimestamp.key.low = *ixdr++;
	verf.adv_int_u = *ixdr++;

	mutex_lock(&authdes_lock);

	/*
	 * Get the conversation key
	 */
	if (fullname_rcvd) {	/* ADN_FULLNAME */
		netobj	pkey;
		char	*pkey_data;

		sessionkey = &cred->adc_fullname.key;
		if (! getpublickey_long(cred->adc_fullname.name, &pkey_data)) {
			__msgout("_svcauth_des: no public key for",
				cred->adc_fullname.name);
			mutex_unlock(&authdes_lock);
			trace1(TR___svcauth_des, 1);
			return (AUTH_BADCRED); /* no key */
		}
		pkey.n_bytes = pkey_data;
		pkey.n_len = strlen(pkey_data) + 1;
		if (key_decryptsession_pk(cred->adc_fullname.name, &pkey,
				sessionkey) < 0) {
		__msgout("_svcauth_des: key_decryptsessionkey failed for",
				cred->adc_fullname.name);
			mutex_unlock(&authdes_lock);
			trace1(TR___svcauth_des, 1);
		        free(pkey_data);
			return (AUTH_BADCRED);	/* key not found */
		}
		free(pkey_data);
	} else { /* ADN_NICKNAME */
		sid = cred->adc_nickname;
		if (sid >= AUTHDES_CACHESZ) {
			__msgout("_svcauth_des:", "bad nickname");
			mutex_unlock(&authdes_lock);
			trace1(TR___svcauth_des, 1);
			return (AUTH_BADCRED);	/* garbled credential */
		}
		sessionkey = &_rpc_authdes_cache[sid].key;
	}

	/*
	 * Decrypt the timestamp
	 */
	cryptbuf[0] = verf.adv_xtimestamp;
	if (fullname_rcvd) {	/* ADN_FULLNAME */
		cryptbuf[1].key.high = cred->adc_fullname.window;
		cryptbuf[1].key.low = verf.adv_winverf;
		ivec.key.high = ivec.key.low = 0;
		status = cbc_crypt((char *)sessionkey, (char *)cryptbuf,
			2 * sizeof (des_block), DES_DECRYPT | DES_HW,
			(char *)&ivec);
	} else {
		status = ecb_crypt((char *)sessionkey, (char *)cryptbuf,
			sizeof (des_block), DES_DECRYPT | DES_HW);
	}
	if (DES_FAILED(status)) {
		__msgout("_svcauth_des: decryption failure for",
			fullname_rcvd ? cred->adc_fullname.name :
			_rpc_authdes_cache[sid].rname);
		mutex_unlock(&authdes_lock);
		trace1(TR___svcauth_des, 1);
		return (AUTH_FAILED);	/* system error */
	}

	/*
	 * XDR the decrypted timestamp
	 */
	ixdr = (__u32 *)cryptbuf;
	timestamp.tv_sec = IXDR_GET_LONG(ixdr);
	timestamp.tv_usec = IXDR_GET_LONG(ixdr);

	/*
	 * Check for valid credentials and verifiers.
	 * They could be invalid because the key was flushed
	 * out of the cache, and so a new session should begin.
	 * Be sure and send AUTH_REJECTED{CRED, VERF} if this is the case.
	 */
	{
		struct timeval current;
		int	nick;
		int	winverf;

		if (fullname_rcvd) {
			window = IXDR_GET_U_LONG(ixdr);
			winverf = IXDR_GET_U_LONG(ixdr);
			if (winverf != window - 1) {
				__msgout("_svcauth_des: corrupted window from",
					cred->adc_fullname.name);
				mutex_unlock(&authdes_lock);
				trace1(TR___svcauth_des, 1);
				/* garbled credential or invalid secret key */
				return (AUTH_BADCRED);
			}
			sid = cache_spot(sessionkey, cred->adc_fullname.name,
					&timestamp);
			if (sid < 0) {
			__msgout("_svcauth_des: replayed credential from",
					cred->adc_fullname.name);
				mutex_unlock(&authdes_lock);
				trace1(TR___svcauth_des, 1);
				return (AUTH_REJECTEDCRED);	/* replay */
			}
			nick = 0;
		} else {	/* ADN_NICKNAME */
			window = _rpc_authdes_cache[sid].window;
			nick = 1;
			if (_rpc_authdes_cache[sid].rname == NULL) {
			  syslog(LOG_ERR, "invalid sid=%d", sid);
			  mutex_unlock(&authdes_lock);
			  trace1(TR___svcauth_des, 1);
			  return AUTH_REJECTEDCRED;
			}
		}

		if (timestamp.tv_usec >= USEC_PER_SEC) {
		__msgout("_svcauth_des: invalid timestamp received from",
				fullname_rcvd ? cred->adc_fullname.name :
					_rpc_authdes_cache[sid].rname);
			/* cached out (bad key), or garbled verifier */
			mutex_unlock(&authdes_lock);
			trace1(TR___svcauth_des, 1);
			return (nick ? AUTH_REJECTEDCRED : AUTH_BADVERF);
		}
		if (nick && BEFORE(&timestamp,
				&_rpc_authdes_cache[sid].laststamp)) {
__msgout("_svcauth_des: timestamp is earlier than the one previously seen from",
			fullname_rcvd ? cred->adc_fullname.name :
				_rpc_authdes_cache[sid].rname);
			mutex_unlock(&authdes_lock);
			trace1(TR___svcauth_des, 1);
			return (AUTH_REJECTEDCRED);	/* replay */
		}
		(void) gettimeofday(&current, (struct timezone *)NULL);
		current.tv_sec -= window;	/* allow for expiration */
		if (!BEFORE(&current, &timestamp)) {
		__msgout("_svcauth_des: timestamp expired for",
				fullname_rcvd ? cred->adc_fullname.name :
					_rpc_authdes_cache[sid].rname);
			/* replay, or garbled credential */
			mutex_unlock(&authdes_lock);
			trace1(TR___svcauth_des, 1);
			return (nick ? AUTH_REJECTEDCRED : AUTH_BADCRED);
		}
	}

	/*
	 * Set up the reply verifier
	 */
	verf.adv_nickname = sid;

	/*
	 * xdr the timestamp before encrypting
	 */
	ixdr = (__u32 *)cryptbuf;
	IXDR_PUT_LONG(ixdr, timestamp.tv_sec - 1);
	IXDR_PUT_LONG(ixdr, timestamp.tv_usec);

	/*
	 * encrypt the timestamp
	 */
	status = ecb_crypt((char *)sessionkey, (char *)cryptbuf,
				sizeof (des_block), DES_ENCRYPT | DES_HW);
	if (DES_FAILED(status)) {
		__msgout("_svcauth_des:", "encryption failure");
		mutex_unlock(&authdes_lock);
		trace1(TR___svcauth_des, 1);
		return (AUTH_FAILED);	/* system error */
	}
	verf.adv_xtimestamp = cryptbuf[0];

	/*
	 * Serialize the reply verifier, and update rqst
	 */
	ixdr = (__u32 *)msg->rm_call.cb_verf.oa_base;
	*ixdr++ = verf.adv_xtimestamp.key.high;
	*ixdr++ = verf.adv_xtimestamp.key.low;
	*ixdr++ = verf.adv_int_u;

	rqst->rq_xprt->xp_verf.oa_flavor = AUTH_DES;
	rqst->rq_xprt->xp_verf.oa_base = msg->rm_call.cb_verf.oa_base;
	rqst->rq_xprt->xp_verf.oa_length =
		(char *)ixdr - msg->rm_call.cb_verf.oa_base;

	/*
	 * We succeeded, commit the data to the cache now and
	 * finish cooking the credential.
	 */
	entry = &_rpc_authdes_cache[sid];
	entry->laststamp = timestamp;
	cache_ref(sid);
	if (cred->adc_namekind == ADN_FULLNAME) {
		cred->adc_fullname.window = window;
		cred->adc_nickname = sid;	/* save nickname */
		if (entry->rname != NULL) {
			mem_free(entry->rname, strlen(entry->rname) + 1);
		}
		entry->rname =
			(char *)mem_alloc((u_int)strlen(cred->adc_fullname.name)
					+ 1);
		if (entry->rname != NULL) {
			(void) strcpy(entry->rname, cred->adc_fullname.name);
		} else {
			__msgout("_svcauth_des:", "out of memory");
		}
		entry->key = *sessionkey;
		entry->window = window;
		/* mark any cached cred invalid */
		invalidate(entry->localcred);
	} else { /* ADN_NICKNAME */
		/*
		 * nicknames are cooked into fullnames
		 */
		cred->adc_namekind = ADN_FULLNAME;
		cred->adc_fullname.name = entry->rname;
		cred->adc_fullname.key = entry->key;
		cred->adc_fullname.window = entry->window;
	}
	mutex_unlock(&authdes_lock);
	trace1(TR___svcauth_des, 1);
	return (AUTH_OK);	/* we made it! */
}


/*
 * Initialize the cache
 */
static void
cache_init()
{
	register int i;

/* LOCK HELD ON ENTRY: authdes_lock */

	trace1(TR_cache_init, 0);
	assert(MUTEX_HELD(&authdes_lock));
	_rpc_authdes_cache = (struct cache_entry *)
		mem_alloc(sizeof (struct cache_entry) * AUTHDES_CACHESZ);
	memset((char *)_rpc_authdes_cache, 0,
		sizeof (struct cache_entry) * AUTHDES_CACHESZ);

	authdes_lru = (short *)mem_alloc(sizeof (short) * AUTHDES_CACHESZ);
	/*
	 * Initialize the lru list
	 */
	for (i = 0; i < AUTHDES_CACHESZ; i++) {
		authdes_lru[i] = i;
	}
	trace1(TR_cache_init, 1);
}

/*
 * Find the lru victim
 */
static short
cache_victim()
{

/* LOCK HELD ON ENTRY: authdes_lock */

	trace1(TR_cache_victim, 0);
	assert(MUTEX_HELD(&authdes_lock));
	trace1(TR_cache_victim, 1);
	return (authdes_lru[AUTHDES_CACHESZ-1]);
}

/*
 * Note that sid was referenced
 */
static void
cache_ref(sid)
	register short sid;
{
	register int i;
	register short curr;
	register short prev;

/* LOCK HELD ON ENTRY: authdes_lock */

	trace1(TR_cache_ref, 0);
	assert(MUTEX_HELD(&authdes_lock));
	prev = authdes_lru[0];
	authdes_lru[0] = sid;
	for (i = 1; prev != sid; i++) {
		curr = authdes_lru[i];
		authdes_lru[i] = prev;
		prev = curr;
	}
	trace1(TR_cache_ref, 1);
}

/*
 * Find a spot in the cache for a credential containing
 * the items given. Return -1 if a replay is detected, otherwise
 * return the spot in the cache.
 */
static short
cache_spot(des_block *key, char *name, struct timeval *timestamp)
{
	register struct cache_entry *cp;
	register int i;
	register __u32 hi;
	static short dummy;

/* LOCK HELD ON ENTRY: authdes_lock */

	trace1(TR_cache_spot, 0);
	assert(MUTEX_HELD(&authdes_lock));
	hi = key->key.high;
	for (cp = _rpc_authdes_cache, i = 0; i < AUTHDES_CACHESZ; i++, cp++) {
		if (cp->key.key.high == hi &&
		    cp->key.key.low == key->key.low &&
		    cp->rname != NULL &&
		    memcmp(cp->rname, name, strlen(name) + 1) == 0) {
			if (BEFORE(timestamp, &cp->laststamp)) {
				svcauthdes_stats.ncachereplays++;
				trace1(TR_cache_spot, 1);
				return (-1);	/* replay */
			}
			svcauthdes_stats.ncachehits++;
			trace1(TR_cache_spot, 1);
			return (i);
			/* refresh */
		}
	}
	svcauthdes_stats.ncachemisses++;
	dummy = cache_victim();
	trace1(TR_cache_spot, 1);
	return (dummy);	/* new credential */
}


/*
 * Local credential handling stuff.
 * NOTE: bsd unix dependent.
 * Other operating systems should put something else here.
 */
#define	UNKNOWN 	-2	/* grouplen, if cached cred is unknown user */
#define	INVALID		-1 	/* grouplen, if cache entry is invalid */

#ifndef NO_ANK_FIX
struct au_cred {
	struct authunix_parms au;
	gid_t groups[NGROUPS];	/* cached groups */
};

/*
 * Map a des credential into a unix cred.
 * We cache the credential here so the application does
 * not have to make an rpc call every time to interpret
 * the credential.
 */
authdes_des2unix (struct authdes_cred *adc, struct authunix_parms ** auc)
{
	unsigned sid;
	register int i;
	uid_t i_uid;	
	gid_t i_gid;
	int i_grouplen;
	struct au_cred *cred;
	gid_t groups[NGROUPS];

	sid = adc->adc_nickname;
	if (sid >= AUTHDES_CACHESZ) {
		__msgout("invalid nickname", "");
		return (0);
	}
	cred = (struct au_cred *)_rpc_authdes_cache[sid].localcred;
	if (cred == NULL) {
		cred = (struct au_cred *)mem_alloc(sizeof(struct au_cred));
		_rpc_authdes_cache[sid].localcred = (char *)cred;
		cred->au.aup_len = INVALID;
		cred->au.aup_gids = cred->groups;
		cred->au.aup_machname = NULL;
	}
	if (cred->au.aup_len == INVALID) {
		/*
		 * not in cache: lookup
		 */
		if (!netname2user(adc->adc_fullname.name, &i_uid, &i_gid, 
			&i_grouplen, groups))
		{
			char host[256];
			static int friends_initted = 0;
			static struct friend {
				struct friend * next;
				char   friend[0];
			} * friends = NULL;

			__inline__ int is_friend(char * host) {
				struct friend * fr;
				for (fr = friends; fr; fr = fr->next) {
					if (strcmp(fr->friend, host) == 0)
						return TRUE;
				}
				return FALSE;
			}


			if (!friends_initted) {
				FILE * fp;
				friends_initted = 1;
				fp = fopen("/etc/hosts.friendly", "r");
				if (fp) {
					while (fgets(host, sizeof(host), fp)) {
						int len = strlen(host);
						struct friend * fr;
						if (len <= 1 || host[0] == '#')
							continue;
						host[len-1] = 0;
						fr = malloc(sizeof(struct friend) + len);
						if (!fr)
							continue;
						memcpy(fr->friend, host, len);
						fr->next = friends;
						friends = fr;
					}
					fclose(fp);
				}
			}
			if (!netname2host(adc->adc_fullname.name, host, sizeof(host))
			   || !is_friend(host)) {
				cred->au.aup_len = UNKNOWN;	/* mark as lookup up, but not found */
				return (0);
			}
			i_uid = i_gid = 0;
			i_grouplen = 0;
			if (cred->au.aup_machname)
				mem_free(cred->au.aup_machname, strlen(cred->au.aup_machname)+1);
			cred->au.aup_machname = mem_alloc(strlen(host)+1);
			if (cred->au.aup_machname)
				strcpy(cred->au.aup_machname, host);
		}
		__msgout("missed ucred cache", "");
		cred->au.aup_uid = i_uid;
		cred->au.aup_gid = i_gid;
		cred->au.aup_len = i_grouplen;
		for (i = i_grouplen - 1; i >= 0; i--)
			cred->groups[i] = groups[i];
	} else if (cred->au.aup_len == UNKNOWN) {
		/*
		 * Already lookup up, but no match found
		 */	
		return (0);
	}

	*auc = &cred->au;
	return (1);
}

static void
invalidate(char * cred)
{
	struct au_cred * auc = (struct au_cred *)cred;
	if (auc == NULL) {
		return;
	}
	auc->au.aup_len = INVALID;
	if (auc->au.aup_machname)
		mem_free(auc->au.aup_machname, strlen(auc->au.aup_machname)+1);
	auc->au.aup_machname = NULL;
}

#else

struct bsdcred {
	uid_t uid;		/* cached uid */
	gid_t gid;		/* cached gid */
	short grouplen;	/* length of cached groups */
	short groups[NGROUPS];	/* cached groups */
};

static void
invalidate(cred)
	char *cred;
{
	trace1(TR_invalidate, 0);
	if (cred == NULL) {
		trace1(TR_invalidate, 1);
		return;
	}
	((struct bsdcred *)cred)->grouplen = INVALID;
	trace1(TR_invalidate, 1);
}
#endif

static void
__msgout(str, strarg)
	const char *str;
	const char *strarg;
{
	trace1(TR___msgout, 0);
#ifdef	KERNEL
		printf("%s %s\n", str, strarg);
#else
		(void) syslog(LOG_WARNING, "%s %s", str, strarg);
#endif
	trace1(TR___msgout, 1);
}
