#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>

#include <dlfcn.h>
#include <tiuser.h>

#include "ti_int.h"

struct ti_device
{
	struct ti_device	*next;
	char 			*name;
	struct ti_module	*module;
	int			(*open)(int oflag);
	int			(*sync)(int fd);
};

struct ti_module
{
	struct ti_module	*next;
	char 			*name;
	int			(*sync)(int fd);
};



static struct ti_device * ti_dev_list = NULL;
static struct ti_module * ti_mod_list = NULL;

static FILE *tlidev_fp;

/* Open /etc/tli_devices */

static int set_tlidev(void)
{
	FILE *fp;
	if (tlidev_fp) {
		rewind(tlidev_fp);
		return 0;
	}
	tlidev_fp = fopen("/etc/tli_devices", "r");
	if (!tlidev_fp)
		return -1;
	return 0;
}

/* Close /etc/tli_devices */

static int end_tlidev(void)
{
	if (tlidev_fp) {
		fclose(tlidev_fp);
		tlidev_fp = NULL;
	}
	return 0;
}

/* Read /etc/tli_devices */

static int get_tlidev(char *dev, char *modname, char *path)
{
	char buf[1024];
	char *name0;
	char *p;

	if (!tlidev_fp)
		return -1;
	while (fgets(buf, sizeof(buf), tlidev_fp) != NULL) {
		p = strrchr(buf, '\n');
		if (p)
			*p = 0;
		p = buf;
		while (*p == ' ' || *p == '\t')
			p++;
		if (*p == '#')
			continue;
		name0 = p;
		while (*p != ' ' && *p != '\t' && *p)
			p++;
		if (*p == 0)
			continue;
		*p++ = 0;
		while (*p == ' ' || *p == '\t')
			p++;
		if (*p == 0)
			continue;
		if (path)
			strncpy(path, p, 1024);
		if (dev)
			strncpy(dev, name0, 1024);
		if (modname) {
			char *q;
			strncpy(modname, p, 1024);
			q = strrchr(p, '/');
			if (q) {
				*q = 0;
				strncpy(modname, q+1, 1024);
			}
		}
		return 0;
	}

	return -1;
}

static int lookup_devname(char *name, char *modname, char *path)
{
	char namebuf[1024];
	if (set_tlidev() < 0)
		return -1;
	while (get_tlidev(namebuf, modname, path) == 0) {
		if (strcmp(name, namebuf) == 0) {
			end_tlidev();
			return 0;
		}
	}
	end_tlidev();
	return -1;
}

static int lookup_modname(char *modname, char *path)
{
	char namebuf[1024];
	if (set_tlidev() < 0)
		return -1;
	while (get_tlidev(NULL, namebuf, path) == 0) {
		if (strcmp(modname, namebuf) == 0) {
			end_tlidev();
			return 0;
		}
	}
	end_tlidev();
	return -1;
}

int __lt_register_module(char *name, int (*sync)(int))
{
	struct ti_module *mod = malloc(sizeof(*mod));
	if (!mod)
		return -1;
	mod->name = malloc(strlen(name) + 1);
	if (!mod->name) {
		free(mod);
		return -1;
	}
	strcpy(mod->name, name);
	mod->sync = sync;
	mod->next = ti_mod_list;
	ti_mod_list = mod;
	return 0;
}

int __lt_register_dev(char *modname, char *name, int (*open)(int))
{
	struct ti_device *dev;
	struct ti_module *mod;
	for (mod = ti_mod_list; mod; mod=mod->next)
		if (strcmp(mod->name, modname) == 0)
			goto found;
	return -1;

found:
	dev = malloc(sizeof(*dev));
	if (!dev)
		return -1;
	dev->name = malloc(strlen(name) + 1);
	if (!dev->name) {
		free(dev);
		return -1;
	}
	strcpy(dev->name, name);
	dev->open = open;
	dev->next = ti_dev_list;
	ti_dev_list = dev;
	dev->module = mod;
	return 0;
}

static struct ti_device *lookup_dev(char *name)
{
	struct ti_device *dev;
	for (dev=ti_dev_list; dev; dev=dev->next) {
		if (strcmp(dev->name, name) == 0)
			return dev;
	}
	return NULL;
}

static struct ti_module *lookup_mod(char *name)
{
	struct ti_module *mod;
	for (mod=ti_mod_list; mod; mod=mod->next) {
		if (strcmp(mod->name, name) == 0)
			return mod;
	}
	return NULL;
}



int t_open(char *path, int oflag, struct t_info *info)
{
	int fd;
	struct ti_device *dev;
	char	modname[1024];
	char	modpath[1024];
	void	*dl;
	int	(*init)(void);
	int	retry=0;

	DEBUG("t_open [%s]", path);
	
	if ((dev = lookup_dev(path)) != NULL)
		goto found;

	if (retry || lookup_devname(path, NULL, modpath) < 0) {
		errno = ENODEV;
		t_errno = TSYSERR;
		return -1;
	}
	
	dl = dlopen(modpath, RTLD_LAZY);
	if (!dl) {
	        syslog(LOG_ERR, "ti_dev: dlopen: %s", dlerror());
		errno = ENODEV;
		t_errno = TSYSERR;
		return -1;
	}

	init = (int (*)())dlsym(dl, "module_init");
	if (!init || init() < 0) {
	        syslog(LOG_ERR, "ti_dev: module_init: %s", dlerror());
		errno = ENODEV;
		t_errno = TSYSERR;
		return -1;
	}
	if ((dev = lookup_dev(path)) != NULL)
		goto found;

	errno = ENODEV;
	t_errno = TSYSERR;
	return -1;

found:
	fd = dev->open(oflag);
	if (fd < 0) {
		t_errno = TSYSERR;
		return -1;
	}
	if (dev->module->sync(fd) < 0) {
		int errno_saved = errno;
		close(fd);
		errno = errno_saved;
		return -1;
	}
	if (info)
		memcpy(info, &__lt_table[fd]->info, sizeof(*info));
	return fd;
}

char *t_device(int fd)
{
	struct lt_desc *lt = lt_lookup(fd);

	if (!lt)
		return NULL;

	TRACE("t_device: fd=%d st=%d", lt->fd, lt->tstate);
	
	return lt->dev;
}

int t_sync(int fd)
{
	struct ti_module *mod;
	char	modname[1024];
	char	modpath[1024];
	void	*dl;
	int	(*init)(void);
	int	state;
	
	DEBUG("t_sync [%d]", fd);

	for (mod=ti_mod_list; mod; mod=mod->next)
		if ((state = mod->sync(fd)) >= 0)
			return state;

	if (set_tlidev() < 0) {
		errno = ENODEV;
		t_errno = TSYSERR;
		return -1;
	}

	while (get_tlidev(NULL, modname, modpath) == 0) {

		if (lookup_mod(modname))
			continue;

		dl = dlopen(modpath, RTLD_LAZY);
		if (!dl)
			continue;

		init = (int (*)())dlsym(dl, "module_init");
		if (!init || init() < 0)
			continue;
		mod = lookup_mod(modname);
		if (!mod)
			continue;
		for (mod=ti_mod_list; mod; mod=mod->next) {
			if ((state = mod->sync(fd)) >= 0) {
				end_tlidev();
				return state;
			}
		}
	}
	end_tlidev();
	t_errno = TSYSERR;
	errno = ENODEV;
	return -1;
}
