#include <stdio.h>
#include <tiuser.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/in6.h>
#include <linux/ipv6.h>
#include <netinet/in.h>
#include <sys/un.h>

#include "ti_int.h"
#include "ti_sock.h"

static const struct in6_addr in6addr_any;

static int
inet6_sockaddr2netbuf(struct sockaddr_un *sun, int alen, struct netbuf *addr)
{
	int i;
	struct sockaddr_in6 *sin = (struct sockaddr_in6*)sun;
	if (addr->maxlen < sizeof(*sin))
		return -1;
	addr->len = sizeof(*sin);
	memcpy(addr->buf, sun, sizeof(*sin));
	return 0;
}

static int
inet6_netbuf2sockaddr(struct sockaddr_un *sun, int *alenp, struct netbuf *addr)
{
	struct sockaddr_in6 *sin = (struct sockaddr_in6*)sun;
	if (addr->len != sizeof(*sin))
		return -1;
	*alenp = sizeof(*sin);
	memcpy(sun, addr->buf, sizeof(*sin));
	return 0;
}

int inet6_optmgmt(struct lt_desc *lt, struct t_optmgmt *req,
		  struct t_optmgmt *ret)
{
	struct msghdr msg;
	struct cmsghdr *cmsg, *out;
	int omaxlen = 0;
	int olen;

	msg.msg_control = req->opt.buf;
	msg.msg_controllen = req->opt.len;

	if (ret) {
		out = (struct cmsghdr*)ret->opt.buf;
		omaxlen = ret->opt.maxlen;
		olen = 0;
	}

	if (req->flags != T_NEGOTIATE)
	{
		t_errno = TNOTSUPPORT;
		return -1;
	}

	for (cmsg=CMSG_FIRSTHDR(&msg); cmsg;
	     cmsg=CMSG_NXTHDR(&msg, cmsg)) {
		int len = cmsg->cmsg_len - sizeof(struct cmsghdr);

		if (setsockopt(lt->fd, cmsg->cmsg_level, cmsg->cmsg_type,
			       CMSG_DATA(cmsg), len) < 0) {
			t_errno = TSYSERR;
			return -1;
		}
		if (olen + cmsg->cmsg_len < omaxlen) {
			*out = *cmsg;
			if (getsockopt(lt->fd, out->cmsg_level,
				       out->cmsg_type, CMSG_DATA(out),
				       &len) < 0)
				continue;
			out->cmsg_len = len + sizeof(*out);
			olen += CMSG_ALIGN(out->cmsg_len);
			out = (struct cmsghdr*)((char*)out+CMSG_ALIGN(out->cmsg_len));
		}
	}
	if (ret)
		ret->opt.len = olen;
	
	return 0;
}

int inet6_laddr(struct lt_desc *lt, int on)
{
	struct sockaddr_in6 *sin = (struct sockaddr_in6 *)lt->bnd.addr.buf;
	if (sin->sin6_family == AF_INET6 &&
	    memcmp(&sin->sin6_addr, &in6addr_any, sizeof(in6addr_any)))
		return -1;
	if (setsockopt(lt->fd, SOL_IPV6, IPV6_PKTINFO, &on, sizeof(on)) < 0) {
		t_errno = TBADOPT;
		return -1;
	}

	return 0;
}

int inet6_rcvladdr(struct lt_desc *lt, struct netbuf *opt, struct netbuf *laddr)
{
	struct msghdr msg;
	struct cmsghdr *cmsg;

	msg.msg_control = opt->buf;
	msg.msg_controllen = opt->len;

	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg;
	     cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_IPV6)
			continue;
		if (cmsg->cmsg_type == IPV6_PKTINFO &&
		    cmsg->cmsg_len >= sizeof(struct cmsghdr) + sizeof(struct in6_pktinfo)) {
			struct in6_addr *a = &((struct in6_pktinfo*)CMSG_DATA(cmsg))->ipi6_addr;

			/* What a pity, IPv6 cannot supply good
			   address for reply
			 */
			if ((a->s6_addr32[0]&htonl(0xFF000000)) == htonl(0xFF000000)) {
				memset(&((struct sockaddr_in6*)laddr->buf)->sin6_addr, 0,
				       sizeof(struct in6_addr));
				return 0;
			}
			memcpy(&((struct sockaddr_in6*)laddr->buf)->sin6_addr, a,
			       sizeof(struct in6_addr));
			return 0;
		}
	}
	return -1;
}

int inet6_sndladdr(struct lt_desc *lt, struct netbuf *opt, struct netbuf *laddr)
{
	int optptr = CMSG_ALIGN(opt->len);
	struct cmsghdr *cmsg=(struct cmsghdr*)(opt->buf+optptr);
	struct sockaddr_in6 *sin = (struct sockaddr_in6*)laddr->buf;

	if (opt->maxlen >=  optptr + sizeof(struct cmsghdr)+sizeof(struct in6_pktinfo)) {
		cmsg->cmsg_len = sizeof(struct cmsghdr)+sizeof(struct in6_pktinfo);
		opt->len += cmsg->cmsg_len;
		cmsg->cmsg_level = SOL_IPV6;
		cmsg->cmsg_type = IPV6_PKTINFO;
		memset(CMSG_DATA(cmsg), 0, sizeof(struct in6_pktinfo));
		memcpy(&((struct in6_pktinfo*)CMSG_DATA(cmsg))->ipi6_addr,
		       &sin->sin6_addr, sizeof(struct in6_addr));
		return 0;
	}
	return -1;
}

int inet6_autobind(struct lt_desc *lt, struct t_bind *bnd)
{
	struct sockaddr_in6 sin;
	struct sockaddr_un sun;
	int	addr_len = sizeof(sin);
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	memset(&sin, 0, sizeof(sin));
	sin.sin6_family = AF_INET6;
 
	if (bind(lt->fd, (struct sockaddr*)&sin, sizeof(sin)) < 0)
	{
		t_errno = TSYSERR;
		return -1;
	}
	if (getsockname(lt->fd, (struct sockaddr*)&sun, &addr_len) < 0)
	{
		t_errno = TSYSERR;
		return -1;
	}
	if (sops->sockaddr2netbuf(&sun, addr_len, &lt->bnd.addr) < 0)
	{
		t_errno = TSYSERR;
		return -1;
	}
#ifdef DEBUGGING
	DEBUG("inet6 autobind %d", lt->fd);
	sops->printaddr(&lt->bnd.addr);
#endif	
	if (bnd && bnd->qlen) {
		listen(lt->fd, bnd->qlen);
		lt->bnd.qlen = bnd->qlen;
	} else
		lt->bnd.qlen = 0;
	return 0;
}

static int inet6_printaddr(struct netbuf *addr)
{
	struct sockaddr_in6 *sin = (struct sockaddr_in6*)addr->buf;
	if (addr->len != sizeof(*sin) ||
	    sin->sin6_family != AF_INET6) {
		syslog(LOG_DEBUG, "bad inet6 address\n");
	}
	syslog(LOG_DEBUG, "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x.%04x\n",
	       ntohs(sin->sin6_addr.s6_addr16[0]), 
	       ntohs(sin->sin6_addr.s6_addr16[1]), 
	       ntohs(sin->sin6_addr.s6_addr16[2]), 
	       ntohs(sin->sin6_addr.s6_addr16[3]), 
	       ntohs(sin->sin6_addr.s6_addr16[4]), 
	       ntohs(sin->sin6_addr.s6_addr16[5]), 
	       ntohs(sin->sin6_addr.s6_addr16[6]), 
	       ntohs(sin->sin6_addr.s6_addr16[7]), 
	       ntohs(sin->sin6_port));
	return 0;
}

struct sock_ops tcpv6_sops = {
	AF_INET6,
	SOCK_STREAM,
	IPPROTO_TCP,
	inet6_sockaddr2netbuf,
	inet6_netbuf2sockaddr,
	NULL,
	inet6_printaddr
};

struct sock_ops udpv6_sops = {
	AF_INET6,
	SOCK_DGRAM,
	IPPROTO_UDP,
	inet6_sockaddr2netbuf,
	inet6_netbuf2sockaddr,
	NULL,
	inet6_printaddr
};
