#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <tiuser.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#ifdef INET6
#include <linux/in6.h>
#endif

#include "ti_int.h"
#include "ti_sock.h"

extern struct proto_ops sock_udp_ops;
extern struct proto_ops sock_tcp_ops;
#ifdef INET6
extern struct proto_ops sock_udpv6_ops;
extern struct proto_ops sock_tcpv6_ops;
#endif
extern struct proto_ops sock_unix_dg_ops;
extern struct proto_ops sock_unix_vc_ops;

struct t_info tcp_info =
{
	sizeof(struct sockaddr_in),
	0,
	0,
	1,
	-2,
	-2,
	T_COTS_ORD
};

struct t_info udp_info =
{
	sizeof(struct sockaddr_in),
	64,
	1024*8+1024,
	-2,
	-2,
	-2,
	T_CLTS
};

#ifdef INET6
struct t_info tcpv6_info =
{
	sizeof(struct sockaddr_in6),
	0,
	0,
	1,
	-2,
	-2,
	T_COTS_ORD
};

struct t_info udpv6_info =
{
	sizeof(struct sockaddr_in6),
	64,
	1024*8+1024,
	-2,
	-2,
	-2,
	T_CLTS
};
#endif

struct t_info unix_vc_info =
{
	sizeof(struct sockaddr_un),
	128,
	0,
	-2,
	-2,
	-2,
	T_COTS_ORD
};

struct t_info unix_dg_info =
{
	sizeof(struct sockaddr_un),
	128,
	8*1024+1024,
	-2,
	-2,
	-2,
	T_CLTS
};


static int tcp_open(int oflags)
{
	return socket(AF_INET, SOCK_STREAM, 0);
}

static int udp_open(int oflags)
{
	return socket(AF_INET, SOCK_DGRAM, 0);
}

#ifdef INET6
static int tcpv6_open(int oflags)
{
	return socket(AF_INET6, SOCK_STREAM, 0);
}

static int udpv6_open(int oflags)
{
	return socket(AF_INET6, SOCK_DGRAM, 0);
}
#endif

static int icmp_open(int oflags)
{
	return socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
}

static int rawip_open(int oflags)
{
	return socket(AF_INET, SOCK_RAW, IPPROTO_IP);
}

static int unix_stream_open(int oflags)
{
	return socket(AF_UNIX, SOCK_STREAM, 0);
}

static int unix_dgram_open(int oflags)
{
	return socket(AF_UNIX, SOCK_DGRAM, 0);
}


int sock_bind(struct lt_desc *lt, struct t_bind *bnd)
{
	struct sockaddr_un sun;
	int addr_len;
	int on = 1;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

#ifdef DEBUGGING
	DEBUG("t_bind %d", lt->fd);
	sops->printaddr(&bnd->addr);
#endif	
	if (sops->netbuf2sockaddr(&sun, &addr_len, &bnd->addr) < 0) {
		ERROR("bad bind addr");
		return -1;
	}
	if (bind(lt->fd, (struct sockaddr*)&sun, addr_len) < 0)
	{
		DEBUG("bind failed: %m");
		if (errno == EADDRINUSE)
			t_errno = TNOADDR;
		else
			t_errno = TSYSERR;
		return -1;
	}
	if (bnd->qlen > 0)
		listen(lt->fd, bnd->qlen);
	netbuf_copy(&lt->bnd.addr, &bnd->addr);
	lt->bnd.qlen = bnd->qlen;
	return 0;
}


static int sock_sync(int fd)
{
	struct sockaddr_un saddr;
	int addr_size = sizeof(saddr);
	int retval;
	int socktype, sockfam; 
	int ssize = sizeof (socktype);
	struct lt_desc *lt;
	struct t_info *info;
	struct sock_ops *sops;
	struct proto_ops *ops;
	char *dev;

	if (fd < 0 || fd >= 256) {
		t_errno = TBADF;
		return -1;
	}

	retval = getsockname(fd, (struct sockaddr *)&saddr, &addr_size);
	if (retval) {
		t_errno = TBADF;
		return -1;
	}
	sockfam = saddr.sun_family;

	if (getsockopt(fd, SOL_SOCKET, SO_TYPE,
		       (char *) &socktype, &ssize) < 0) {
		t_errno = TBADF;
		return -1;
	}

	if (sockfam == AF_INET && socktype == SOCK_STREAM) {
		info = &tcp_info;
		ops = &sock_tcp_ops;
		sops = &tcp_sops;
		dev = "/dev/tcp";
	} else if (sockfam == AF_INET && socktype == SOCK_DGRAM) {
		info = &udp_info;
		ops = &sock_udp_ops;
		sops = &udp_sops;
		dev = "/dev/udp";
#ifdef INET6
	} else if (sockfam == AF_INET6 && socktype == SOCK_STREAM) {
		info = &tcpv6_info;
		ops = &sock_tcpv6_ops;
		sops = &tcpv6_sops;
		dev = "/dev/tcpv6";
	} else if (sockfam == AF_INET6 && socktype == SOCK_DGRAM) {
		info = &udpv6_info;
		ops = &sock_udpv6_ops;
		sops = &udpv6_sops;
		dev = "/dev/udpv6";
#endif
	} else if (sockfam == AF_UNIX && socktype == SOCK_STREAM) {
		info = &unix_vc_info;
		ops = &sock_unix_vc_ops;
		sops = &unix_stream_sops;
		dev = "/dev/ticotsord";
	} else if (sockfam == AF_UNIX && socktype == SOCK_DGRAM) {
		info = &unix_dg_info;
		ops = &sock_unix_dg_ops;
		sops = &unix_dgram_sops;
		dev = "/dev/ticlts";
	} else {
		fprintf(stderr, "sock_sync: sockfam=%d type=%d\n", sockfam, socktype);
		t_errno = TBADF;
		return -1;
	}

	lt = __lt_table[fd];
	if (lt) {
		if (lt->ops != ops || lt->priv != (void*)sops ||
		    memcmp(&lt->info, info, sizeof(*info))) {
			__t_unsync(fd);
			lt = NULL;
		}
	}

	if (!lt) {
		lt = __lt_table[fd] = malloc(sizeof(*lt));
		if (!lt) {
			errno = ENOMEM;
			t_errno = TSYSERR;
			return -1;
		}
		memset(lt, 0, sizeof(*lt));

		lt->tstate = T_UNINIT;

		memcpy(&lt->info, info, sizeof(*info));

		lt->dev = dev;
		lt->ops = ops;
		lt->priv = (void*)sops;
		lt->fd = fd;
		lt->peercred.pid = 0;
		lt->peercred.uid = -1;
		lt->peercred.gid = -1;
		lt->bnd.addr.maxlen = info->addr;
		lt->bnd.addr.len = 0;
		lt->bnd.addr.buf = malloc(info->addr);
		if (!lt->bnd.addr.buf) {
			__t_unsync(fd);
			t_errno = TSYSERR;
			errno = ENOMEM;
			return -1;
		}

		lt->peer.addr.maxlen = info->addr;
		lt->peer.addr.len = 0;
		lt->peer.addr.buf = malloc(info->addr);
		if (!lt->peer.addr.buf) {
			__t_unsync(fd);
			t_errno = TSYSERR;
			errno = ENOMEM;
			return -1;
		}
	}

	info = &lt->info;
	sops = (struct sock_ops *)lt->priv;

	if (sops->sockaddr2netbuf(&saddr, addr_size, &lt->bnd.addr) < 0)
	{
		__t_unsync(fd);
		t_errno = TBADADDR;
		return -1;
	}

	if (sockfam == AF_INET) {
		struct sockaddr_in *sin = (struct sockaddr_in*)&saddr;
		if (sin->sin_port)
			lt->tstate = T_IDLE;
		else
			lt->tstate = T_UNBND;
#ifdef INET6
	} else if (sockfam == AF_INET6) {
		struct sockaddr_in6 *sin = (struct sockaddr_in6*)&saddr;
		if (sin->sin6_port)
			lt->tstate = T_IDLE;
		else
			lt->tstate = T_UNBND;
#endif		
	} else {
		if (lt->bnd.addr.len)
			lt->tstate = T_IDLE;
		else
			lt->tstate = T_UNBND;
	}

	if (socktype != SOCK_STREAM)
		return lt->tstate;

	if (lt->tstate == T_DATAXFER ||
	    lt->tstate == T_INREL ||
	    lt->tstate == T_OUTREL)
		return lt->tstate;

	retval = getpeername(fd, (struct sockaddr *)&saddr, &addr_size);
	if (retval)
		return lt->tstate;

	if (sops->sockaddr2netbuf(&saddr, addr_size, &lt->peer.addr) < 0)
	{
		__t_unsync(fd);
		t_errno = TBADADDR;
		return -1;
	}

	lt->tstate = T_DATAXFER;

        {
		int olen=sizeof(lt->peercred);
		if (getsockopt(lt->fd, SOL_SOCKET, SO_PEERCRED,
			       &lt->peercred, &olen) < 0)
			syslog(LOG_ERR, "SO_PEERCRED : %m");
        }
	
	return lt->tstate;
}

struct acc_elt *sock_accept(struct lt_desc *lt, struct t_call *call)
{
	struct sockaddr_un sun;
	int addr_len;
	struct acc_elt *acc;
	int newfd;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	if ((newfd = accept(lt->fd, (struct sockaddr*)&sun, &addr_len)) < 0)
	{
		t_errno = TSYSERR;
		return NULL;
	}
	acc = malloc(sizeof(struct acc_elt));
	if (!acc)
	{
		close(newfd);
		t_errno = TSYSERR;
		return NULL;
	}
	if (t_sync(newfd) < 0)
	{
		close(newfd);
		return NULL;
	}

	acc->lt = __lt_table[newfd];
	acc->lt->flags = lt->flags;

	call->opt.len = 0;
	if (call->opt.maxlen > 0  && sops->makeoptions)
		sops->makeoptions(acc->lt, &call->opt);

	return acc;
}

int sock_connect(struct lt_desc *lt, struct t_call *snd, struct t_call *rcv)
{
	struct sockaddr_un sun;
	int addr_len;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	if (!snd || sops->netbuf2sockaddr(&sun, &addr_len, &snd->addr) < 0) {
		t_errno = TBADADDR;
		return -1;
	}
	if (snd->opt.len) {
		t_errno = TBADOPT;
		return -1;
	}
	if (snd->udata.len) {
		t_errno = TBADDATA;
		return -1;
	}

	netbuf_copy(&lt->peer.addr, &snd->addr);

restart:
	if (lt->tevent) {
		t_errno = TLOOK;
		return -1;
	}
	if (connect(lt->fd, (struct sockaddr*)&sun, addr_len) < 0) {
		if (errno == EINPROGRESS || errno == EALREADY) {
			lt->tstate = T_OUTCON;
			t_errno = TNODATA;
			return -1;
		}
		if (errno == EINTR)
			goto restart;
		if (errno != EISCONN) {
			t_errno = TSYSERR;
			return -1;
		}
	}

	if (rcv)
	{
		rcv->udata.len = 0;
		rcv->opt.len = 0;
		netbuf_copy(&rcv->addr, &snd->addr);
	}
	return 0;
}

int sock_rcvconnect(struct lt_desc *lt, struct t_call *rcv)
{
	struct sockaddr_un sun;
	int addr_len;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	sops->netbuf2sockaddr(&sun, &addr_len, &lt->peer.addr);

restart:
	if (lt->tevent) {
		t_errno = TLOOK;
		return -1;
	}
	if (connect(lt->fd, (struct sockaddr*)&sun, addr_len) < 0) {
		if (errno == EINPROGRESS || errno == EALREADY) {
			t_errno = TNODATA;
			return -1;
		}
		if (errno == EINTR)
			goto restart;
		if (errno != EISCONN) {
			t_errno = TSYSERR;
			return -1;
		}
	}

	if (rcv)
	{
		rcv->udata.len = 0;
		rcv->opt.len = 0;
		netbuf_copy(&rcv->addr, &lt->peer.addr);
	}
	return 0;
}

int sock_dgram_recv(struct lt_desc *lt, struct t_unitdata *data, int *flags)
{
	int len;
	void *ptr;
	struct sockaddr_un sun;
	struct iovec iov;
	struct msghdr msg;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	iov.iov_base = data->udata.buf;
	iov.iov_len = data->udata.maxlen;
	msg.msg_name = (char*)&sun;
	msg.msg_namelen = sizeof(sun);
	msg.msg_control = data->opt.buf;
	msg.msg_controllen = data->opt.maxlen;
	msg.msg_flags = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

restart:
	len = recvmsg(lt->fd, &msg, 0);
	if (len >= 0) {
		if (sops->sockaddr2netbuf(&sun, msg.msg_namelen, &data->addr)<0)
			syslog(LOG_ERR, "bad sockaddr");
		data->opt.len = msg.msg_controllen;
		if (msg.msg_flags & MSG_CTRUNC)
			data->opt.len = 0;
		*flags = 0;
		data->udata.len = len;
		if (len == 0)
			ERROR("sock_dgram_recv: %d len==0 impossible", lt->fd);
		return len;
	}
	if (errno == EINTR)
		goto restart;
	if (errno == EAGAIN || errno == ECONNREFUSED) {
		t_errno = TNODATA;
		return -1;
	}
	t_errno = TSYSERR;
	return -1;
}

int sock_stream_recv(struct lt_desc *lt, struct t_unitdata *data, int *flags)
{
	int len;

	if (data->udata.maxlen == 0) {
		t_errno = TBADDATA;
		return -1;
	}

again:
	len = recv(lt->fd, data->udata.buf, data->udata.maxlen, 0);
	if (len > 0)
	{
		data->udata.len = len;
		*flags = 0;
		TRACE("sock_stream_recv: %d %d bytes", lt->fd, len);
		return len;
	}

	if (len == 0) {
		lt->tevent |= T_ORDREL;
		t_errno = TLOOK;
		TRACE("sock_stream_recv: %d eof", lt->fd);
		return -1;
	}
	if (errno == EINTR)
		goto again;

	if (errno == EAGAIN) {
		t_errno = TNODATA;
		TRACE("sock_stream_recv: %d again", lt->fd);
		return -1;
	}
	if (errno == ECONNRESET || errno == ECONNREFUSED ||
	    errno == EPIPE) {
		lt->tevent |= T_DISCONNECT;
		t_errno = TLOOK;
		DEBUG("sock_stream_recv: %d %m", lt->fd);
		return -1;
	}
	ERROR("sock_stream_recv: %d bad error %m", lt->fd);
	t_errno = TSYSERR;
	return -1;
}

int sock_dgram_send(struct lt_desc *lt, struct t_unitdata *data, int flags)
{
	int len;
	struct sockaddr_un sun;
	struct iovec iov;
	struct msghdr msg;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	if (data->udata.len == 0) {
		t_errno = TBADDATA;
		return -1;
	}
	if (sops->netbuf2sockaddr(&sun, &msg.msg_namelen, &data->addr) < 0) {
		t_errno = TBADADDR;
		return -1;
	}
	iov.iov_base = data->udata.buf;
	iov.iov_len = data->udata.len;
	msg.msg_name = (char*)&sun;
	msg.msg_control = data->opt.buf;
	msg.msg_controllen = data->opt.len;
	msg.msg_flags = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

restart:
	len = sendmsg(lt->fd, &msg, 0);
	if (len >= 0)
		return len;
	if (errno == EINTR)
		goto restart;
	if (errno == EAGAIN) {
		t_errno = TFLOW;
		return -1;
	}
	t_errno = TSYSERR;
	return -1;
}


int sock_stream_send(struct lt_desc *lt, struct t_unitdata *data, int flags)
{
	int msgflags = 0;
	int len;
	int bytes = data->udata.len;
	void *ptr;

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_DATAXFER && lt->tstate != T_INREL) {
		t_errno = TOUTSTATE;
		return -1;
	}
	if (bytes == 0) {
		t_errno = TBADDATA;
		return -1;
	}
	if (flags & T_EXPEDITED)
		msgflags = MSG_OOB;
	ptr=data->udata.buf;
	len=0;
	while (bytes) {
		int chunk;
		chunk = send(lt->fd, ptr, bytes, msgflags);
		if (chunk > 0) {
			bytes -= chunk;
			len += chunk;
			continue;
		}
		if (chunk == 0) {
			/* Impossible .... */
			ERROR("send_vc on fd=%d returnd zero", lt->fd);
			if (len)
				return len;
			t_errno = TBADDATA;
			return -1;
		}
		if (errno == EINTR)
			continue;
		if (errno == EAGAIN) {
			if (len)
				return len;
			t_errno = TFLOW;
			return -1;
		}
		if (errno == ECONNRESET || errno == ECONNREFUSED ||
		    errno == EPIPE) {
			lt->tevent |= T_DISCONNECT;
			if (len)
				return len;
			t_errno = TLOOK;
			return -1;
		}
		if (len)
			return len;
		t_errno = TSYSERR;
		return -1;
	}
	return len;
}

int sock_shutdown(struct lt_desc *lt)
{
	if (shutdown(lt->fd, 1) == 0)
		return 0;
	t_errno = TSYSERR;
	return -1;
}

static int sock_auth(struct lt_desc *lt, int on)
{
	if (on)
		lt->flags |= LT_RECV_CRED;
	else
		lt->flags &= ~LT_RECV_CRED;
	if (setsockopt(lt->fd, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on)) < 0) {
		t_errno = TBADOPT;
		return -1;
	}
	return 0;
}

static int sock_rcvauth(struct lt_desc *lt, struct netbuf *opt, struct ucred *cred)
{
	struct msghdr msg;
	struct cmsghdr *cmsg;

	msg.msg_control = opt->buf;
	msg.msg_controllen = opt->len;

	for (cmsg=CMSG_FIRSTHDR(&msg); cmsg;
	     cmsg=CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level == SOL_SOCKET &&
		    cmsg->cmsg_type == SCM_CREDENTIALS &&
		    cmsg->cmsg_len >= sizeof(*cmsg) + sizeof(struct ucred)) {
			struct ucred *mcred = (struct ucred*)CMSG_DATA(cmsg);
			if (mcred->pid == 0 || mcred->uid == (uid_t)-1)
				return -1;
			memcpy(cred, mcred, sizeof(*cred));
			return 0;
		}
	}
	return -1;
}

int sock_sndauth(struct lt_desc *lt, struct netbuf *opt, struct ucred *cred)
{
	int optptr = CMSG_ALIGN(opt->len);
	struct cmsghdr *cmsg=(struct cmsghdr*)(opt->buf+optptr);

	if (opt->maxlen >=  optptr + sizeof(struct cmsghdr)+sizeof(struct ucred)) {
		cmsg->cmsg_len = sizeof(struct cmsghdr)+sizeof(struct ucred);
		opt->len += cmsg->cmsg_len;
		cmsg->cmsg_level = SOL_SOCKET;
		cmsg->cmsg_type = SCM_CREDENTIALS;
		memcpy(CMSG_DATA(cmsg), cred, sizeof(*cred));
		return 0;
	}
	return -1;
}

struct proto_ops sock_tcp_ops = {
	tcp_open,
	sock_sync,
	NULL,
	sock_bind,
	inet_autobind,
	sock_accept,
	sock_connect,
	sock_rcvconnect,
	sock_stream_recv,
	sock_stream_send,
	sock_shutdown,
	inet_optmgmt,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

struct proto_ops sock_udp_ops = {
	udp_open,
	sock_sync,
	NULL,
	sock_bind,
	inet_autobind,
	sock_accept,
	sock_connect,
	sock_rcvconnect,
	sock_dgram_recv,
	sock_dgram_send,
	sock_shutdown,
	inet_optmgmt,

	inet_laddr,
	inet_rcvladdr,
	inet_sndladdr,
	NULL,
	NULL,
	NULL
};

#ifdef INET6
struct proto_ops sock_tcpv6_ops = {
	tcpv6_open,
	sock_sync,
	NULL,
	sock_bind,
	inet6_autobind,
	sock_accept,
	sock_connect,
	sock_rcvconnect,
	sock_stream_recv,
	sock_stream_send,
	sock_shutdown,
	inet6_optmgmt,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL

};

struct proto_ops sock_udpv6_ops = {
	udpv6_open,
	sock_sync,
	NULL,
	sock_bind,
	inet6_autobind,
	sock_accept,
	sock_connect,
	sock_rcvconnect,
	sock_dgram_recv,
	sock_dgram_send,
	sock_shutdown,
	inet6_optmgmt,

	inet6_laddr,
	inet6_rcvladdr,
	inet6_sndladdr,
	NULL,
	NULL,
	NULL

};
#endif

struct proto_ops sock_unix_vc_ops = {
	unix_stream_open,
	sock_sync,
	NULL,
	sock_bind,
	unix_autobind,
	sock_accept,
	sock_connect,
	sock_rcvconnect,
	sock_stream_recv,
	sock_stream_send,
	sock_shutdown,
	unix_optmgmt,

	NULL,
	NULL,
	NULL,
	sock_auth,
	sock_rcvauth,
	sock_sndauth
};

struct proto_ops sock_unix_dg_ops = {
	unix_dgram_open,
	sock_sync,
	NULL,
	sock_bind,
	unix_autobind,
	sock_accept,
	sock_connect,
	sock_rcvconnect,
	sock_dgram_recv,
	sock_dgram_send,
	sock_shutdown,
	unix_optmgmt,

	NULL,
	NULL,
	NULL,
	sock_auth,
	sock_rcvauth,
	sock_sndauth
};


int module_init(void)
{
	static int initted;
	if (initted)
		return -1;
	__lt_register_module("tisock.so", sock_sync);
	__lt_register_dev("tisock.so", "/dev/tcp", tcp_open);
	__lt_register_dev("tisock.so", "/dev/udp", udp_open);
#ifdef INET6
	__lt_register_dev("tisock.so", "/dev/tcpv6", tcpv6_open);
	__lt_register_dev("tisock.so", "/dev/udpv6", udpv6_open);
#endif
	__lt_register_dev("tisock.so", "/dev/ticotsord", unix_stream_open);
	__lt_register_dev("tisock.so", "/dev/ticlts", unix_dgram_open);
	return 0;
}
