#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <tiuser.h>

#include "ti_int.h"

struct lt_desc *__lt_table[256];

int __t_unsync(int fd)
{
	struct lt_desc *lt = __lt_lookup(fd);

	if (lt->ops->unsync)
		lt->ops->unsync(lt);

	__lt_table[fd] = NULL;

	if (lt->bnd.addr.buf)
		free(lt->bnd.addr.buf);
	if (lt->peer.addr.buf)
		free(lt->peer.addr.buf);
	if (lt->acc_q)
		__lt_purge_acc_q(lt);
	free(lt);
	return 0;
}

int t_unbind(int fd)
{
	struct lt_desc *lt = __lt_lookup(fd);

	NOTIMPLEMENTED("t_unbind fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->tstate != T_IDLE) {
		t_errno = TOUTSTATE;
		return -1;
	}
	t_errno = TNOTSUPPORT;
	return -1;
}

void netbuf_copy(struct netbuf *dst, struct netbuf *src)
{
	if (src->buf) {
		if (dst->buf) {
			dst->len = src->len;
			if (dst->len > dst->maxlen)
				dst->len = dst->maxlen;
			memcpy(dst->buf, src->buf, dst->len);
		}
	} else if (dst->buf)
		dst->len = 0;
}

int t_close(int fd)
{
	__t_unsync(fd);
	
	TRACE("t_close fd=%d", fd);

	if (close(fd) < 0) {
		t_errno = TSYSERR;
		return -1;
	}
	return 0;
}

int t_getinfo(int fd, struct t_info *info)
{
	struct lt_desc *lt = __lt_lookup(fd);
	memcpy(info, &lt->info, sizeof(*info));
	return 0;
}

int t_bind(int fd, struct t_bind *bnd, struct t_bind *res)
{
	int err;
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_bind: fd=%d st=%d", fd, lt->tstate);

	if (lt->tstate != T_UNBND) {
		TRACE("t_bind: TOUTSTATE");
		t_errno = TOUTSTATE;
		return -1;
	}

	if (bnd && bnd->addr.len)
		err = lt->ops->bind(lt, bnd);
	else
		err = lt->ops->autobind(lt, bnd);

	if (err < 0)
		return -1;
	
	lt->tstate = T_IDLE;
	if (res) {
		netbuf_copy(&res->addr, &lt->bnd.addr);
		res->qlen = lt->bnd.qlen;
	}
	return 0;
}

int t_getprotaddr(int fd, struct t_bind *bnd, struct t_bind *peer)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_getprotaddr fd=%d st=%d", lt->fd, lt->tstate);

	if (bnd) {
		netbuf_copy(&bnd->addr, &lt->bnd.addr);
		bnd->qlen = lt->bnd.qlen;
	}
	if (peer) {
		netbuf_copy(&peer->addr, &lt->peer.addr);
		peer->qlen = 0;
	}
}

int t_getname(int fd, struct t_bind *res, int what)
{
	if (what == LOCALNAME)
		return t_getprotaddr(fd, res, NULL);
	else
		return t_getprotaddr(fd, NULL, res);
}

int t_getstate(int fd)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_getstate fd=%d st=%d", lt->fd, lt->tstate);
	return lt->tstate;
}


int t_listen(int fd, struct t_call *call)
{
	static long cur_seq_number=1;
	struct lt_desc *lt = __lt_lookup(fd);
	struct acc_elt *acc;

	TRACE("t_listen fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TBADF;
		return -1;
	}

	if (lt->tstate != T_IDLE && lt->tstate != T_INCON) {
		t_errno = TOUTSTATE;
		return -1;
	}

	acc = lt->ops->accept(lt, call);
	if (!acc)
		return -1;
	call->sequence = acc->sequence = cur_seq_number++;

	lt_acc_link(lt, acc);

	netbuf_copy(&call->addr, &acc->lt->peer.addr);

	lt->acc_q_len++;
	lt->tstate = T_INCON;

	lt->tevent &= ~T_LISTEN;
	return 0;
}


int t_accept(int fd, int accfd, struct t_call *call)
{
	struct lt_desc *newlt;
	struct lt_desc *lt = __lt_lookup(fd);
	struct lt_desc *acclt = __lt_lookup(accfd);
	struct acc_elt *acc;
	int newfd;

	TRACE("t_accept fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_INCON) {
		t_errno = TOUTSTATE;
		return -1;
	}

	if (acclt->info.servtype != lt->info.servtype ||
	    acclt->ops != lt->ops) {
		t_errno = TNOTSUPPORT /*TPROVMISMATCH*/;
		return -1;
	}

	if (acclt != lt) {
		if (acclt->bnd.qlen) {
			t_errno = TNOTSUPPORT /* TRESQLEN*/;
			return -1;
		}
		if (acclt->tstate != T_UNBND && acclt->tstate != T_IDLE) {
			t_errno = TOUTSTATE;
			return -1;
		}
	}

	for (acc = lt->acc_q; acc; acc=acc->next) {
		if (acc->sequence == call->sequence)
			goto found;
	}

	t_errno = TBADSEQ;
	return -1;


found:
	if (acclt == lt && lt->acc_q_len != 1) {
		t_errno = TOUTSTATE /*TINDOUT */;
		return -1;
	}
	if (!--lt->acc_q_len)
		lt->tstate = T_IDLE;

	lt_acc_unlink(lt, acc);

	__t_unsync(accfd);

	__lt_table[accfd] = acc->lt;
	__lt_table[acc->lt->fd] = NULL;

	if (accfd != fd) {
		dup2(acc->lt->fd, accfd);
		close(acc->lt->fd);
	} else
		dup2(acc->lt->fd, fd);

	acc->lt->fd = accfd;

	free(acc);
	return 0;
}


int t_connect(int fd, struct t_call *snd, struct t_call *rcv)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_connect fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_IDLE) {
		t_errno = TOUTSTATE;
		return -1;
	}

	if (lt->ops->connect(lt, snd, rcv) < 0)
		return -1;

	lt->tstate = T_DATAXFER;

	return 0;
}

int t_rcvconnect(int fd, struct t_call *rcv)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_rcvconnect fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_OUTCON) {
		t_errno = TOUTSTATE;
		return -1;
	}

	if (lt->ops->rcvconnect(lt, rcv) < 0)
		return -1;

	lt->tstate = T_DATAXFER;

	lt->tevent &= ~T_CONNECT;
	return 0;
}

int t_rcvudata(int fd, struct t_unitdata *data, int *flags)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_rcvudata fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype != T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_IDLE) {
		t_errno = TOUTSTATE;
		return -1;
	}
	if (!data) {
		t_errno = TBADDATA;
		return -1;
	}
	return lt->ops->recv(lt, data, flags) >= 0 ? 0 : -1;
}

int t_rcv(int fd, char *buf, unsigned bytes, int *flags)
{
	struct lt_desc *lt = __lt_lookup(fd);
	struct t_unitdata udata;
	int err;

	TRACE("t_rcv fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_DATAXFER && lt->tstate != T_OUTREL) {
		ERROR("t_rcv: %d bad state %d", lt->fd, lt->tstate);
		t_errno = TOUTSTATE;
		return -1;
	}
	if (lt->tevent & ~T_DATA) {
		ERROR("t_rcv: %d tlook %d %08x", lt->fd, lt->tstate, lt->tevent);
		t_errno = TLOOK;
		return -1;
	}
	lt->tevent &= ~T_DATA;
	memset (&udata, 0, sizeof(udata));
	udata.udata.maxlen = bytes;
	udata.udata.len = 0;
	udata.udata.buf = buf;
	return lt->ops->recv(lt, &udata, flags);
}

int t_rcvuderr(int fd, struct t_uderr *uderr)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_rcvuderr fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype != T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}

	lt->tevent &= ~T_UDERR;
	t_errno = TNOTSUPPORT;
	return -1;
}


int t_rcvdis(int fd, struct t_discon *discon)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_rcvdis fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate == T_UNBND || lt->tstate == T_IDLE
	    || lt->tstate == T_INCON)
	{
		t_errno = TOUTSTATE;
		return -1;
	}
	if (!(lt->tevent & T_DISCONNECT)) {
		t_errno = TNODIS;
		return -1;
	}
	lt->tevent &= ~T_DISCONNECT;
	if (discon)
		discon->udata.len = 0;
	
	return 0;
}

int t_rcvrel(int fd)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_rcvrel: fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_DATAXFER && lt->tstate != T_OUTREL)
	{
		t_errno = TOUTSTATE;
		return -1;
	}

	if (!(lt->tevent & T_ORDREL)) {
		ERROR("t_rcvrel: %d no rel", lt->fd);
		t_errno = TNOREL;
		return -1;
	}
	lt->tevent &= ~T_ORDREL;

	if (lt->tstate == T_DATAXFER)
		lt->tstate = T_INREL;
	else
		lt->tstate = T_UNINIT;
	
	return 0;
}


int t_snd(int fd, char *buf, unsigned bytes, int flags)
{
	struct lt_desc *lt = __lt_lookup(fd);
	struct t_unitdata data;

	TRACE("t_snd fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_DATAXFER && lt->tstate != T_INREL) {
		t_errno = TOUTSTATE;
		return -1;
	}
	if (lt->tevent) {
		t_errno = TLOOK;
		return -1;
	}
	memset(&data, 0, sizeof(data));
	data.udata.len = data.udata.maxlen = bytes;
	data.udata.buf = buf;
	return lt->ops->send(lt, &data, flags);
}

int t_sndudata(int fd, struct t_unitdata *data /*, int flags*/)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_sndudata fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype != T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_IDLE) {
		t_errno = TOUTSTATE;
		return -1;
	}
	if (!data) {
		t_errno = TBADDATA;
		return -1;
	}
	return lt->ops->send(lt, data, 0/*flags*/) >= 0 ? 0 : -1;
}

int t_sndrel(int fd)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_sndrel: fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype != T_COTS_ORD) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_DATAXFER && lt->tstate != T_INREL) {
		t_errno = TOUTSTATE;
		return -1;
	}
	if (lt->tevent) {
		t_errno = TLOOK;
		return -1;
	}

	if (lt->ops->shutdown(lt) < 0)
		return -1;

	if (lt->tstate == T_DATAXFER)
		lt->tstate = T_OUTREL;
	else
		lt->tstate = T_UNINIT;
	return 0;
}


int t_snddis(int fd, struct t_call *call)
{
	struct lt_desc *lt = __lt_lookup(fd);
	struct acc_elt *acc;

	TRACE("t_snddis fd=%d st=%d", lt->fd, lt->tstate);

	if (lt->info.servtype == T_CLTS) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate == T_IDLE) {
		t_errno = TOUTSTATE;
		return -1;
	}
	if (lt->tstate == T_OUTCON) {
		t_errno = TNOTSUPPORT;
		return -1;
	}
	if (lt->tstate != T_INCON) {
		t_errno = TNOTSUPPORT;
		return -1;
	}

	for (acc = lt->acc_q; acc; acc=acc->next) {
		if (acc->sequence == call->sequence)
			goto found;
	}

	t_errno = TBADSEQ;
	return -1;

found:
	if (!--lt->acc_q_len)
		lt->tstate = T_IDLE;
	t_close(acc->lt->fd);
	lt_acc_unlink(lt, acc);
	free(acc);
	return 0;
}

int t_optmgmt(int fd, struct t_optmgmt *req, struct t_optmgmt *ret)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("t_optmgmt fd=%d st=%d", lt->fd, lt->tstate);

	if (!req) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->optmgmt(lt, req, ret);
}



int tiopt_laddr(int fd, int on)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("tiopt_laddr fd=%d st=%d on=%d", lt->fd, lt->tstate, on);

	if (!lt->ops->opt_laddr) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->opt_laddr(lt, on);
}

int tiopt_rcvladdr(int fd, struct netbuf *opt, struct netbuf *laddr)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("tiopt_rcvladdr fd=%d st=%d", lt->fd, lt->tstate);

	if (!lt->ops->opt_rcvladdr) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->opt_rcvladdr(lt, opt, laddr);
}

int tiopt_sndladdr(int fd, struct netbuf *opt, struct netbuf *laddr)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("tiopt_sndladdr fd=%d st=%d", lt->fd, lt->tstate);

	if (!lt->ops->opt_sndladdr) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->opt_sndladdr(lt, opt, laddr);
}

int tiopt_auth(int fd, int on)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("tiopt_auth fd=%d st=%d %d", lt->fd, lt->tstate, on);

	if (!lt->ops->opt_auth) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->opt_auth(lt, on);
}

int tiopt_rcvauth(int fd, struct netbuf *opt, struct ucred *cred)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("tiopt_rcvauth fd=%d st=%d", lt->fd, lt->tstate);

	if (!lt->ops->opt_rcvauth) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->opt_rcvauth(lt, opt, cred);
}

int tiopt_sndauth(int fd, struct netbuf *opt, struct ucred *cred)
{
	struct lt_desc *lt = __lt_lookup(fd);

	TRACE("tiopt_sndauth fd=%d st=%d", lt->fd, lt->tstate);

	if (!lt->ops->opt_sndauth) {
		t_errno = TBADOPT;
		return -1;
	}
	return lt->ops->opt_sndauth(lt, opt, cred);
}
